# coding=utf-8
"""
Copyright 2013 LinkedIn Corp. All rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from collections import defaultdict
import glob
import logging
import os
import re
from naarad.graphing.plot_data import PlotData as PD
import naarad.utils
import naarad.httpdownload
import naarad.naarad_constants as CONSTANTS
import datetime
import heapq
from luminol import anomaly_detector, correlator

logger = logging.getLogger('naarad.metrics.metric')


class Metric(object):

  def __init__(self, metric_type, infile_list, hostname, aggr_metrics, output_directory, resource_path, label, ts_start, ts_end,
               rule_strings, important_sub_metrics, anomaly_detection_metrics, **other_options):
    self.metric_type = metric_type
    self.infile_list = infile_list
    self.hostname = hostname
    self.outdir = output_directory
    self.resource_path = resource_path
    self.resource_directory = os.path.join(self.outdir, self.resource_path)
    self.label = label
    self.ts_start = naarad.utils.get_standardized_timestamp(ts_start, None)
    self.ts_end = naarad.utils.get_standardized_timestamp(ts_end, None)
    self.calc_metrics = None
    self.precision = None
    self.sep = ','
    self.titles_string = None
    self.ylabels_string = None
    self.csv_files = []
    self.plot_files = []
    self.stats_files = []
    self.important_stats_files = []
    self.percentiles_files = []
    self.column_csv_map = {}
    self.csv_column_map = {}
    self.sub_metric_description = defaultdict(lambda: 'None')  # the description of the submetrics.
    self.sub_metric_unit = defaultdict(lambda: 'None')      # the unit of the submetrics.  The plot will have the Y-axis being: Metric name (Unit)
    self.important_sub_metrics = important_sub_metrics
    self.sla_list = []  # TODO : remove this once report has grading done in the metric tables
    self.sla_map = defaultdict(lambda: defaultdict(lambda: defaultdict(None)))
    self.calculated_stats = {}
    self.calculated_percentiles = {}
    self.summary_stats_list = CONSTANTS.DEFAULT_SUMMARY_STATS
    self.summary_stats = defaultdict(dict)
    self.status = CONSTANTS.OK
    self.ignore = False
    self.timezone = "PDT"
    self.options = None
    self.sub_metrics = None   # users can specify what sub_metrics to process/plot;
    self.groupby = None
    self.summary_charts = []
    self.aggregation_granularity = aggr_metrics   # default is none
    # Leave the flag here for the future use to control summary page
    self.summary_html_content_enabled = True
    self.anomaly_detection_metrics = anomaly_detection_metrics
    self.anomalies = {}
    for (key, val) in rule_strings.iteritems():
      naarad.utils.set_sla(self, self.label, key, val)
    if other_options:
      for (key, val) in other_options.iteritems():
        setattr(self, key, val)
      if not self.titles_string:
        self.titles_string = self.columns
      if self.columns:
        self.columns = self.columns.split()
      if self.groupby:
        self.groupby = self.groupby.split()

      self.titles = dict(zip(self.columns, self.titles_string.split(','))) if self.columns and self.titles_string else None
      self.ylabels = dict(zip(self.columns, self.ylabels_string.split(','))) if self.columns and self.ylabels_string else None

  def name_to_index(self, name):
    index = None
    for i in range(len(self.columns)):
      if name == self.columns[i]:
        index = i + 1
    return index

  def get_groupby_indexes(self, groupby):
    groupby_indexes = []
    for group in groupby:
      if ':' in group.rstrip(':'):
        name, index = group.split(':')
        groupby_indexes.append(index)
      else:
        groupby_indexes.append(self.name_to_index(group.rstrip(':')))
    return groupby_indexes

  def ts_out_of_range(self, timestamp):
    if self.ts_start and timestamp < self.ts_start:
      return True
    elif self.ts_end and timestamp > self.ts_end:
      return True
    return False

  def collect_local(self, infile):
    return os.path.exists(infile)

  def collect(self):
    # self.infile_list can be of several formats: for instance a local dir (e.g., /path/a.log) or an http url;
    # decide the case based on self.infile_list;
    # self.access is optional, can be removed.
    collected_files = []
    for infile in self.infile_list:
      if infile.startswith("http://") or infile.startswith("https://"):
        if naarad.utils.is_valid_url(infile):
          http_download_dir = os.path.join(self.outdir, self.label)
          output_file = naarad.httpdownload.download_url_single(infile, http_download_dir)
          if output_file:
            collected_files.append(output_file)
          else:
            return False
        else:
          logger.error("The given url of {0} is invalid.\n".format(infile))
          return False
      else:
        file_matches = glob.glob(infile)
        if len(file_matches) == 0:
          return False
        for file_name in file_matches:
          if self.collect_local(file_name):
            collected_files.append(file_name)
          else:
            return False
    self.infile_list = collected_files
    return True

  def get_csv(self, column, groupby=None):
    if groupby:
      column = groupby + '.' + column
    if column in self.column_csv_map.keys():
      return self.column_csv_map[column]
    col = naarad.utils.sanitize_string(column)
    csv = os.path.join(self.resource_directory, self.label + '.' + col + '.csv')
    self.csv_column_map[csv] = column
    self.column_csv_map[column] = csv
    return csv

  def get_important_sub_metrics_csv(self):
    csv = os.path.join(self.resource_directory, self.label + '.important_sub_metrics.csv')
    return csv

  def get_stats_csv(self):
    csv = os.path.join(self.resource_directory, self.label + '.stats.csv')
    return csv

  def get_percentiles_csv_from_data_csv(self, data_csv):
    percentile_csv_file = '.'.join(data_csv.split('.')[0:-1]) + '.percentiles.csv'
    return percentile_csv_file

  def get_sla_csv(self):
    csv = os.path.join(self.resource_directory, self.label + '.sla.csv')
    return csv

  def update_summary_stats(self, column):
    for stat in self.summary_stats_list:
      if stat.startswith('p'):
        self.summary_stats[column][stat] = naarad.utils.normalize_float_for_display(self.calculated_percentiles[column][int(stat[1:])])
      else:
        self.summary_stats[column][stat] = naarad.utils.normalize_float_for_display(self.calculated_stats[column][stat])

  def get_aggregation_timestamp(self, timestamp, granularity='second'):
    """
    Return a timestamp from the raw epoch time based on the granularity preferences passed in.

    :param string timestamp: timestamp from the log line
    :param string granularity: aggregation granularity used for plots.
    :return: string aggregate_timestamp: timestamp used for metrics aggregation in all functions
    """
    if granularity is None or granularity.lower() == 'none':
      return int(timestamp), 1
    elif granularity == 'hour':
      return (int(timestamp) / (3600 * 1000)) * 3600 * 1000, 3600
    elif granularity == 'minute':
      return (int(timestamp) / (60 * 1000)) * 60 * 1000, 60
    else:
      return (int(timestamp) / 1000) * 1000, 1

  def aggregate_count_over_time(self, metric_store, groupby_name, aggregate_timestamp):
    """
    Organize and store the count of data from the log line into the metric store by columnm, group name, timestamp

    :param dict metric_store: The metric store used to store all the parsed the log data
    :param string groupby_name: the group name that the log line belongs to
    :param string aggregate_timestamp: timestamp used for storing the raw data. This accounts for aggregation time period
    :return: None
    """
    all_qps = metric_store['qps']
    qps = all_qps[groupby_name]
    if aggregate_timestamp in qps:
      qps[aggregate_timestamp] += 1
    else:
      qps[aggregate_timestamp] = 1
    return None

  def aggregate_values_over_time(self, metric_store, data, groupby_name, column_name, aggregate_timestamp):
    """
    Organize and store the data from the log line into the metric store by metric type, transaction, timestamp

    :param dict metric_store: The metric store used to store all the parsed log data
    :param string data: column data in the log line
    :param string groupby_name: the group that the data belongs to
    :param string column_name: the column name of the data
    :param string aggregate_timestamp: timestamp used for storing the raw data. This accounts for aggregation time period
    :return: None
    """
    # To add overall_summary one
    if self.groupby:
      metric_data = reduce(defaultdict.__getitem__, [column_name, 'Overall_summary', aggregate_timestamp], metric_store)
      metric_data.append(float(data))
    metric_data = reduce(defaultdict.__getitem__, [column_name, groupby_name, aggregate_timestamp], metric_store)
    metric_data.append(float(data))
    return None

  def average_values_for_plot(self, metric_store, data, averaging_factor):
    """
    Create the time series for the various metrics, averaged over the aggregation period being used for plots

    :param dict metric_store: The metric store used to store all the parsed log data
    :param dict data: Dict with all the metric data to be output to csv
    :param float averaging_factor: averaging factor to be used for calculating the average per second metrics
    :return: None
    """
    for column, groups_store in metric_store.items():
      for group, time_store in groups_store.items():
        for time_stamp, column_data in sorted(time_store.items()):
          if column in ['qps']:
            if self.groupby:
              data[self.get_csv(column, group)].append(','.join([str(time_stamp), str(column_data / float(averaging_factor))]))
            else:
              data[self.get_csv(column)].append(','.join([str(time_stamp), str(column_data / float(averaging_factor))]))
          else:
            if self.groupby:
              data[self.get_csv(column, group)].append(','.join([str(time_stamp), str(sum(map(float, column_data)) / float(len(column_data)))]))
            else:
              data[self.get_csv(column)].append(','.join([str(time_stamp), str(sum(map(float, column_data)) / float(len(column_data)))]))
    return None

  def parse(self):
    processed_data = defaultdict(lambda: defaultdict(lambda: defaultdict(list)))
    data = defaultdict(list)
    groupby_idxes = None
    averaging_factor = None
    if self.groupby:
      groupby_idxes = self.get_groupby_indexes(self.groupby)
    for input_file in self.infile_list:
      logger.info("Working on " + input_file)
      timestamp_format = None
      with open(input_file, 'r') as infile:
        for line in infile:
          if self.sep is None or self.sep == '':
            words = line.strip().split()
          else:
            words = line.strip().split(self.sep)
          if len(words) == 0:
            continue
          if len(words) <= len(self.columns):  # NOTE: len(self.columns) is always one less than len(words) since we assume the very first column is timestamp
            logger.warning("WARNING: Number of columns given in config is more than number of columns present in line {0}\n", line)
            continue
          if not timestamp_format or timestamp_format == 'unknown':
            timestamp_format = naarad.utils.detect_timestamp_format(words[0])
          if timestamp_format == 'unknown':
            continue
          ts = naarad.utils.get_standardized_timestamp(words[0], timestamp_format)
          if ts == -1:
            continue
          ts = naarad.utils.reconcile_timezones(ts, self.timezone, self.graph_timezone)
          if self.ts_out_of_range(ts):
            continue
          if self.groupby:
            groupby_names = None
            for index in groupby_idxes:
              if not groupby_names:
                groupby_names = words[index].rstrip(':')
              else:
                groupby_names += '.' + words[index].rstrip(':')
            aggregate_timestamp, averaging_factor = self.get_aggregation_timestamp(ts, self.aggregation_granularity)
            self.aggregate_count_over_time(processed_data, groupby_names, aggregate_timestamp)
            for i in range(len(self.columns)):
              if i + 1 in groupby_idxes:
                continue
              else:
                self.aggregate_values_over_time(processed_data, words[i + 1], groupby_names, self.columns[i], aggregate_timestamp)
          else:
            groupby_names = 'DEFAULT'
            aggregate_timestamp, averaging_factor = self.get_aggregation_timestamp(ts, self.aggregation_granularity)
            self.aggregate_count_over_time(processed_data, groupby_names, aggregate_timestamp)
            for i in range(len(self.columns)):
              self.aggregate_values_over_time(processed_data, words[i + 1], groupby_names, self.columns[i], aggregate_timestamp)
    # Post processing, putting data in csv files
    self.average_values_for_plot(processed_data, data, averaging_factor)
    for csv in data.keys():
      self.csv_files.append(csv)
      with open(csv, 'w') as fh:
        fh.write('\n'.join(sorted(data[csv])))
    self.calc_key_stats(processed_data)
    return True

  def calc_key_stats(self, metric_store):
    """
    Calculate stats such as percentile and mean

    :param dict metric_store: The metric store used to store all the parsed log data
    :return: None
    """
    stats_to_calculate = ['mean', 'std', 'min', 'max']  # TODO: get input from user
    percentiles_to_calculate = range(0, 100, 1)  # TODO: get input from user
    for column, groups_store in metric_store.items():
      for group, time_store in groups_store.items():
        data = metric_store[column][group].values()
        if self.groupby:
          column_name = group + '.' + column
        else:
          column_name = column
        if column.startswith('qps'):
          self.calculated_stats[column_name], self.calculated_percentiles[column_name] = naarad.utils.calculate_stats(data, stats_to_calculate, percentiles_to_calculate)
        else:
          self.calculated_stats[column_name], self.calculated_percentiles[column_name] = naarad.utils.calculate_stats(list(heapq.merge(*data)), stats_to_calculate,
                                                                                                            percentiles_to_calculate)
        self.update_summary_stats(column_name)

  def calculate_stats(self):
    """
    Calculate stats with different function depending on the metric type:
    Data is recorded in memory for base metric type, and use calculate_base_metric_stats()
    Data is recorded in CSV file for other metric types, and use calculate_other_metric_stats()

    """
    metric_type = self.metric_type.split('-')[0]
    if metric_type in naarad.naarad_imports.metric_classes or metric_type in naarad.naarad_imports.aggregate_metric_classes:
      self.calculate_other_metric_stats()
    else:
      self.calculate_base_metric_stats()

  def calculate_base_metric_stats(self):
    stats_to_calculate = ['mean', 'std', 'min', 'max']  # TODO: get input from user
    percentiles_to_calculate = range(0, 101, 1)  # TODO: get input from user
    headers = CONSTANTS.SUBMETRIC_HEADER + ',mean,std,p50,p75,p90,p95,p99,min,max\n'  # TODO: This will be built from user input later on
    metric_stats_csv_file = self.get_stats_csv()
    imp_metric_stats_csv_file = self.get_important_sub_metrics_csv()
    imp_metric_stats_present = False
    metric_stats_present = False
    logger.info("Calculating stats for important sub-metrics in %s and all sub-metrics in %s", imp_metric_stats_csv_file, metric_stats_csv_file)
    with open(metric_stats_csv_file, 'w') as FH:
      with open(imp_metric_stats_csv_file, 'w') as FH_IMP:
        FH.write(headers)
        FH_IMP.write(headers)
        for sub_metric in self.calculated_stats:
          percentile_data = self.calculated_percentiles[sub_metric]
          stats_data = self.calculated_stats[sub_metric]
          csv_data = ','.join([sub_metric] + map(lambda x: str(round(x, 2)), [stats_data['mean'], stats_data['std'],
                                                                              percentile_data[50], percentile_data[75],
                                                                              percentile_data[90], percentile_data[95],
                                                                              percentile_data[99], stats_data['min'],
                                                                              stats_data['max']]))
          FH.write(csv_data + '\n')
          if self.important_sub_metrics is not None and sub_metric in self.important_sub_metrics:
            FH_IMP.write(csv_data + '\n')
        self.stats_files.append(metric_stats_csv_file)
        self.important_stats_files.append(imp_metric_stats_csv_file)
    for column in self.calculated_percentiles:
      csv_file = self.column_csv_map[column]
      percentiles_csv_file = self.get_percentiles_csv_from_data_csv(csv_file)
      percentile_data = self.calculated_percentiles[column]
      with open(percentiles_csv_file, 'w') as FH:
        for percentile in sorted(percentile_data):
          FH.write(str(percentile) + ',' + str(round(percentile_data[percentile], 2)) + '\n')
        self.percentiles_files.append(percentiles_csv_file)

  def calculate_other_metric_stats(self):
    stats_to_calculate = ['mean', 'std', 'min', 'max']  # TODO: get input from user
    percentiles_to_calculate = range(0, 101, 1)  # TODO: get input from user
    headers = CONSTANTS.SUBMETRIC_HEADER + ',mean,std,p50,p75,p90,p95,p99,min,max\n'  # TODO: This will be built from user input later on
    metric_stats_csv_file = self.get_stats_csv()
    imp_metric_stats_csv_file = self.get_important_sub_metrics_csv()
    imp_metric_stats_present = False
    metric_stats_present = False
    logger.info("Calculating stats for important sub-metrics in %s and all sub-metrics in %s", imp_metric_stats_csv_file, metric_stats_csv_file)
    with open(metric_stats_csv_file, 'w') as FH_W:
      with open(imp_metric_stats_csv_file, 'w') as FH_W_IMP:
        for csv_file in self.csv_files:
          data = []
          value_error = False
          if not os.path.getsize(csv_file):
            continue
          column = self.csv_column_map[csv_file]
          percentile_csv_file = self.get_percentiles_csv_from_data_csv(csv_file)
          with open(csv_file, 'r') as FH:
            for line in FH:
              words = line.split(',')
              try:
                data.append(float(words[1]))
              except ValueError:
                if not value_error:
                  logger.error("Cannot convert to float. Some data is ignored in file " + csv_file)
                  value_error = True
                continue
          self.calculated_stats[column], self.calculated_percentiles[column] = naarad.utils.calculate_stats(data, stats_to_calculate, percentiles_to_calculate)
          with open(percentile_csv_file, 'w') as FH_P:
            for percentile in sorted(self.calculated_percentiles[column].iterkeys()):
              FH_P.write("%d, %f\n" % (percentile, self.calculated_percentiles[column][percentile]))
          self.percentiles_files.append(percentile_csv_file)
          self.update_summary_stats(column)
          to_write = [column, self.calculated_stats[column]['mean'], self.calculated_stats[column]['std'], self.calculated_percentiles[column][50],
                      self.calculated_percentiles[column][75], self.calculated_percentiles[column][90], self.calculated_percentiles[column][95],
                      self.calculated_percentiles[column][99], self.calculated_stats[column]['min'], self.calculated_stats[column]['max']]
          to_write = map(lambda x: naarad.utils.normalize_float_for_display(x), to_write)
          if not metric_stats_present:
            metric_stats_present = True
            FH_W.write(headers)
          FH_W.write(','.join(to_write) + '\n')
          # Important sub-metrics and their stats go in imp_metric_stats_csv_file
          sub_metric = column
          if self.metric_type in self.device_types:
            sub_metric = column.split('.')[1]
          if self.important_sub_metrics and sub_metric in self.important_sub_metrics:
            if not imp_metric_stats_present:
              FH_W_IMP.write(headers)
              imp_metric_stats_present = True
            FH_W_IMP.write(','.join(to_write) + '\n')
        if imp_metric_stats_present:
          self.important_stats_files.append(imp_metric_stats_csv_file)
      self.stats_files.append(metric_stats_csv_file)

  def calc(self):
    if not self.calc_metrics:
      return
    calculation_array = self.calc_metrics.split()
    for calculation in calculation_array:
      words = calculation.split('=')
      newmetric = words[0]
      expr = words[1]
      p = re.compile('(\w+)\((.+)\)')
      calc_type = p.match(expr).group(1)
      old_metric = p.match(expr).group(2)
      logger.debug('In calc() : %s %s %s %s', newmetric, expr, old_metric, calc_type)
      if calc_type not in ('rate', 'diff'):
        logger.error("ERROR: Invalid calc_metric type {0} defined in config".format(calc_type))
        continue
      old_metric_csv = self.get_csv(old_metric)
      new_metric_csv = self.get_csv(newmetric)
      self.csv_files.append(new_metric_csv)
      old_val = None
      with open(old_metric_csv, 'r') as FH:
        with open(new_metric_csv, 'w') as NEW_FH:
          for line in FH:
            w = line.split(',')
            ts = w[0]
            val = w[1]
            if not old_val:
              old_ts = ts
              old_val = val
              continue
            if calc_type == 'rate':
              # Multiply rate by 1000 since timestamp is in ms
              ts_diff = naarad.utils.convert_to_unixts(ts) - naarad.utils.convert_to_unixts(old_ts)
              if ts_diff != 0:
                new_metric_val = 1000 * (float(val) - float(old_val)) / ts_diff
              else:
                new_metric_val = 0
                logger.warn("rate calculation encountered zero timestamp difference")
            elif calc_type == 'diff':
              new_metric_val = (float(val) - float(old_val))
            old_ts = ts
            old_val = val
            NEW_FH.write(ts)
            NEW_FH.write(',')
            NEW_FH.write(str(new_metric_val))
            NEW_FH.write('\n')

  def plot_timeseries(self, graphing_library='matplotlib'):
    """
    plot timeseries for sub-metrics
    """
    if self.groupby:
      plot_data = {}
      # plot time series data for submetrics
      for out_csv in sorted(self.csv_files, reverse=True):
        csv_filename = os.path.basename(out_csv)
        transaction_name = ".".join(csv_filename.split('.')[1:-1])
        if transaction_name in self.anomalies.keys():
          highlight_regions = self.anomalies[transaction_name]
        else:
          highlight_regions = None
        # The last element is .csv, don't need that in the name of the chart
        column = csv_filename.split('.')[-2]
        transaction_name = ' '.join(csv_filename.split('.')[1:-2])
        plot = PD(input_csv=out_csv, csv_column=1, series_name=transaction_name + '.' + column,
                  y_label=column + ' (' + self.sub_metric_description[column] + ')', precision=None, graph_height=500, graph_width=1200, graph_type='line',
                  highlight_regions=highlight_regions)
        if transaction_name in plot_data:
          plot_data[transaction_name].append(plot)
        else:
          plot_data[transaction_name] = [plot]
      for transaction in plot_data:
        graphed, div_file = Metric.graphing_modules[graphing_library].graph_data(plot_data[transaction], self.resource_directory, self.resource_path,
                                                                                 self.label + '.' + transaction)
        if graphed:
          self.plot_files.append(div_file)
    else:
      graphed = False
      for out_csv in self.csv_files:
        csv_filename = os.path.basename(out_csv)
        transaction_name = ".".join(csv_filename.split('.')[1:-1])
        if transaction_name in self.anomalies.keys():
          highlight_regions = self.anomalies[transaction_name]
        else:
          highlight_regions = None
        # The last element is .csv, don't need that in the name of the chart
        column = self.csv_column_map[out_csv]
        column = naarad.utils.sanitize_string(column)
        graph_title = '.'.join(csv_filename.split('.')[0:-1])
        if self.sub_metric_description and column in self.sub_metric_description.keys():
          graph_title += ' (' + self.sub_metric_description[column] + ')'
        if self.sub_metric_unit and column in self.sub_metric_unit.keys():
          plot_data = [PD(input_csv=out_csv, csv_column=1, series_name=graph_title, y_label=column + ' (' + self.sub_metric_unit[column] + ')',
                          precision=None, graph_height=600, graph_width=1200, graph_type='line', highlight_regions=highlight_regions)]
        else:
          plot_data = [PD(input_csv=out_csv, csv_column=1, series_name=graph_title, y_label=column, precision=None, graph_height=600, graph_width=1200,
                          graph_type='line', highlight_regions=highlight_regions)]
        graphed, div_file = Metric.graphing_modules[graphing_library].graph_data(plot_data, self.resource_directory, self.resource_path, graph_title)
        if graphed:
          self.plot_files.append(div_file)
    return True

  def check_important_sub_metrics(self, sub_metric):
    """
    check whether the given sub metric is in important_sub_metrics list
    """
    if not self.important_sub_metrics:
      return False
    if sub_metric in self.important_sub_metrics:
      return True
    items = sub_metric.split('.')
    if items[-1] in self.important_sub_metrics:
      return True
    return False

  def plot_cdf(self, graphing_library='matplotlib'):
    """
    plot CDF for important sub-metrics
    """
    graphed = False
    for percentile_csv in self.percentiles_files:
      csv_filename = os.path.basename(percentile_csv)
      # The last element is .csv, don't need that in the name of the chart
      column = self.csv_column_map[percentile_csv.replace(".percentiles.", ".")]
      if not self.check_important_sub_metrics(column):
        continue
      column = naarad.utils.sanitize_string(column)
      graph_title = '.'.join(csv_filename.split('.')[0:-1])
      if self.sub_metric_description and column in self.sub_metric_description.keys():
        graph_title += ' (' + self.sub_metric_description[column] + ')'
      if self.sub_metric_unit and column in self.sub_metric_unit.keys():
        plot_data = [PD(input_csv=percentile_csv, csv_column=1, series_name=graph_title, x_label='Percentiles',
                        y_label=column + ' (' + self.sub_metric_unit[column] + ')', precision=None, graph_height=600, graph_width=1200, graph_type='line')]
      else:
        plot_data = [PD(input_csv=percentile_csv, csv_column=1, series_name=graph_title, x_label='Percentiles', y_label=column, precision=None,
                        graph_height=600, graph_width=1200, graph_type='line')]
      graphed, div_file = Metric.graphing_modules[graphing_library].graph_data_on_the_same_graph(plot_data, self.resource_directory,
                                                                                                 self.resource_path, graph_title)
      if graphed:
        self.plot_files.append(div_file)
    return True

  def graph(self, graphing_library='matplotlib'):
    """
    graph generates two types of graphs
    'time': generate a time-series plot for all submetrics (the x-axis is a time series)
    'cdf': generate a CDF plot for important submetrics (the x-axis shows percentiles)
    """
    logger.info('Using graphing_library {lib} for metric {name}'.format(lib=graphing_library, name=self.label))
    self.plot_cdf(graphing_library)
    self.plot_timeseries(graphing_library)
    return True

  def detect_anomaly(self):
    """
    Detect anomalies in the timeseries data for the submetrics specified in the config file. Identified anomalies are
    stored in self.anomalies as well as written to .anomalies.csv file to be used by the client charting page. Anomaly
    detection uses the luminol library (http://pypi.python.org/pypi/luminol)
    """
    if not self.anomaly_detection_metrics or len(self.anomaly_detection_metrics) <= 0:
      return
    for submetric in self.anomaly_detection_metrics:
      csv_file = self.get_csv(submetric)
      if naarad.utils.is_valid_file(csv_file):
        detector = anomaly_detector.AnomalyDetector(csv_file)
        anomalies = detector.get_anomalies()
        if len(anomalies) <= 0:
          return
        self.anomalies[submetric] = anomalies
        anomaly_csv_file = os.path.join(self.resource_directory, self.label + '.' + submetric + '.anomalies.csv')
        with open(anomaly_csv_file, 'w') as FH:
          for anomaly in anomalies:
            FH.write(",".join([str(anomaly.anomaly_score), str(anomaly.start_timestamp), str(anomaly.end_timestamp), str(anomaly.exact_timestamp)]))
            FH.write('\n')



# coding=utf-8
"""
Copyright 2013 LinkedIn Corp. All rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import sys
import logging
import os
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), 'src')))
from naarad.reporting.diff import Diff
from naarad.reporting.diff import NaaradReport

diff_obj = None


def setup_module():
  global diff_obj
  reports_list = []
  report_name = 'diff'
  output_directory = ''
  resource_directory = ''
  resource_path = 'resource_path'
  naarad_reports = [NaaradReport('/tmp', None), NaaradReport('/tmp', None)]
  diff_obj = Diff(naarad_reports, 'diff', '/tmp', '/tmp', 'resources')


def test_collect_cdf_datasources():
  """
  Test whether collect_cdf_datasources works as expected
  :return: None
  """
  global diff_obj
  diff_obj.reports[0].cdf_datasource = ['a.csv', 'b.csv', 'c.csv']
  diff_obj.reports[1].cdf_datasource = ['a.csv', 'b.csv', 'd.csv']
  return_code = diff_obj.collect_cdf_datasources()
  assert return_code is True
  assert diff_obj.reports[0].cdf_datasource == diff_obj.reports[1].cdf_datasource
  assert len(diff_obj.reports[0].cdf_datasource) == 2
  assert diff_obj.reports[0].cdf_datasource[0] == 'a.csv'
  assert diff_obj.reports[0].cdf_datasource[1] == 'b.csv'
  diff_obj.reports[0].cdf_datasource = []
  diff_obj.reports[1].cdf_datasource = ['a.csv', 'b.csv', 'd.csv']
  return_code = diff_obj.collect_cdf_datasources()
  assert return_code is False



# coding=utf-8
"""
Copyright 2013 LinkedIn Corp. All rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import sys
import uuid
import shutil

# add the path of ~/naarad/src;   the testing py is under ~/naarad/test
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), 'src')))
import naarad.utils
from naarad.metrics.netstat_metric import NetstatMetric

# the temporary directory for testing, will remove it after done.
tmp_dir = ''


def prepare_data():
  """
  Hard code the raw logs and output into files so that netstat metric can pick them up.
  Doing so can remove the dependency on physical logs.
  :return:
  """
  log = []
  log.append('2014-04-14 12:09:01.67581	Active Internet connections (w/o servers)')
  log.append('2014-04-14 12:09:01.67581	Proto Recv-Q Send-Q Local Address               Foreign Address             State       PID/Program name')
  log.append('2014-04-14 12:09:01.67581	tcp        0      500 host1.localdomain.com:43214 web1.remotedomain.com:https ESTABLISHED 4996/firefox')
  log.append('2014-04-14 12:09:01.67581	tcp        120      0 host1.localdomain.com:48860 email.localdomain.com:https ESTABLISHED 4996/firefox')
  log.append('2014-04-14 12:09:03.76251	Active Internet connections (w/o servers)')
  log.append('2014-04-14 12:09:03.76251	Proto Recv-Q Send-Q Local Address               Foreign Address             State       PID/Program name')
  log.append('2014-04-14 12:09:03.76251	tcp        0      200 host1.localdomain.com:43214 web1.remotedomain.com:https ESTABLISHED 4996/firefox')
  log.append('2014-04-14 12:09:03.76251	tcp        330      0 host1.localdomain.com:48860 email.localdomain.com:https ESTABLISHED 4996/firefox')
  log.append('2014-04-14 12:09:05.84302	Active Internet connections (w/o servers)')
  log.append('2014-04-14 12:09:05.84302	Proto Recv-Q Send-Q Local Address               Foreign Address             State       PID/Program name')
  log.append('2014-04-14 12:09:05.84302	tcp        0      345 host1.localdomain.com:43214 web1.remotedomain.com:https ESTABLISHED 4996/firefox')
  log.append('2014-04-14 12:09:05.84302	tcp        440      0 host1.localdomain.com:48860 email.localdomain.com:https ESTABLISHED 4996/firefox')
  log.append('2014-04-14 12:09:07.91455	Active Internet connections (w/o servers)')
  log.append('2014-04-14 12:09:07.91455	Proto Recv-Q Send-Q Local Address               Foreign Address             State       PID/Program name')
  log.append('2014-04-14 12:09:07.91455	tcp        0      0 host1.localdomain.com:43214 web1.remotedomain.com:https ESTABLISHED 4996/firefox')
  log.append('2014-04-14 12:09:07.91455	tcp        1550      0 host1.localdomain.com:48860 email.localdomain.com:https ESTABLISHED 4996/firefox')
  log.append('2014-04-14 12:09:09.98031	Active Internet connections (w/o servers)')
  log.append('2014-04-14 12:09:09.98031	Proto Recv-Q Send-Q Local Address               Foreign Address             State       PID/Program name')
  log.append('2014-04-14 12:09:09.98031	tcp        0      564 host1.localdomain.com:43214 web1.remotedomain.com:https ESTABLISHED 4996/firefox')
  log.append('2014-04-14 12:09:09.98031	tcp        20      0 host1.localdomain.com:48860 email.localdomain.com:https ESTABLISHED 4996/firefox')
  log.append('2014-04-14 12:09:12.05993	Active Internet connections (w/o servers)')
  log.append('2014-04-14 12:09:12.05993	Proto Recv-Q Send-Q Local Address               Foreign Address             State       PID/Program name')
  log.append('2014-04-14 12:09:12.05993	tcp        0      234 host1.localdomain.com:43214 web1.remotedomain.com:https ESTABLISHED 4996/firefox')
  log.append('2014-04-14 12:09:12.05993	tcp        3245      0 host1.localdomain.com:48860 email.localdomain.com:https ESTABLISHED 4996/firefox')

  with open(os.path.join(tmp_dir, 'netstat.tcp.out'), 'w') as fh:
    fh.write('\n'.join(log))


def setup():
  create_tmp_dir()
  prepare_data()


def teardown():
  delete_tmp_dir()


def create_tmp_dir():
  """
  create a unique tmp dir to hold the downloaded local files
  if the tmp_dir grenerated already exists, then simply return
  the user simply try again to generate another unique tmp dir
  :return:
  """
  global tmp_dir
  tmp_dir = os.path.join('./', 'tmp' + '.' + str(uuid.uuid4()))   # ./tmp.'randomstring'
  if not os.path.exists(tmp_dir):
    os.makedirs(tmp_dir)
  else:
    print "the path of %s already exists, please try again." % tmp_dir
    return


def delete_tmp_dir():
  """
  delete the tmp directory
  :return:
  """
  shutil.rmtree(tmp_dir)


def test_netstatmetric():
  """
  First construct a NetstatMetric, then call the parse(), finally check whether the output files are there
  :return:
  """
  # construct a NetstatMetric
  section = 'NETSTAT-host1'
  label = 'NETSTAT-host1'
  hostname = 'localhost'
  resource_path = 'resources'
  rule_strings = {}
  output_directory = tmp_dir
  infile_list = ['netstat.tcp.out']
  other_options = {'connections': 'host1.localdomain.com<->web1.remotedomain.com:https host1:48860<->email', 'processes': '/firefox'}
  ts_start = None
  ts_end = None
  anomaly_detection_metrics = None
  important_sub_metrics = []

  cur_metric = NetstatMetric(section, infile_list, hostname, output_directory, resource_path, label, ts_start, ts_end, rule_strings, important_sub_metrics,
                             anomaly_detection_metrics, **other_options)
  cur_metric.infile_list = [os.path.join(tmp_dir, f) for f in cur_metric.infile_list]

  # create sub-directory of resource_path
  sub_dir = os.path.join(output_directory, resource_path)
  if not os.path.exists(sub_dir):
    os.makedirs(sub_dir)

  cur_metric.parse()

  # check the existance of the output files
  output_files = ['NETSTAT-host1.host1.localdomain.com_43214.web1.remotedomain.com_https.RecvQ.csv',
                  'NETSTAT-host1.host1.localdomain.com_43214.web1.remotedomain.com_https.SendQ.csv',
                  'NETSTAT-host1.host1.localdomain.com_48860.email.localdomain.com_https.RecvQ.csv',
                  'NETSTAT-host1.host1.localdomain.com_48860.email.localdomain.com_https.SendQ.csv']
  for f in output_files:
    file_path = os.path.join(sub_dir, f)
    assert os.path.exists(file_path)

if __name__ == '__main__':
  test_netstatmetric()



# coding=utf-8
"""
Copyright 2013 LinkedIn Corp. All rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import sys
import re
import urllib2
import logging
from HTMLParser import HTMLParser

import naarad.utils

logger = logging.getLogger('naarad.httpdownload')


def handle_single_url(url, outdir, outfile=None):
  """
  Base function which takes a single url, download it to outdir/outfile
  :param str url: a full/absolute url, e.g. http://www.cnn.com/log.zip
  :param str outdir: the absolute local directory. e.g. /home/user1/tmp/
  :param str outfile: (optional) filename stored in local directory. If outfile is not given, extract the filename from url
  :return: the local full path name of downloaded url
  """
  if not url or type(url) != str \
     or not outdir or type(outdir) != str:
      logger.error('passed in parameters %s %s are incorrect.' % (url, outdir))
      return

  if not naarad.utils.is_valid_url(url):
    logger.error("passed in url %s is incorrect." % url)
    return

  if not outfile:
    segs = url.split('/')
    outfile = segs[-1]
    outfile = urllib2.quote(outfile)

  output_file = os.path.join(outdir, outfile)
  if os.path.exists(output_file):
    logger.warn("the %s already exists!" % outfile)

  with open(output_file, "w") as fh:
    try:
      response = urllib2.urlopen(url)
      fh.write(response.read())
    except urllib2.HTTPError:
      logger.error("got HTTPError when retrieving %s" % url)
      return
    except urllib2.URLError:
      logger.error("got URLError when retrieving %s" % url)
      return

  return output_file


def stream_url(url):
  """
  Read response of specified url into memory and return to caller. No persistence to disk.
  :return: response content if accessing the URL succeeds, False otherwise
  """
  try:
    response = urllib2.urlopen(url)
    response_content = response.read()
    return response_content
  except (urllib2.URLError, urllib2.HTTPError) as e:
    logger.error('Unable to access requested URL: %s', url)
    return False


class HTMLLinkExtractor(HTMLParser):
  """
  Helper class to parse the html file returned. It extracts href links into links[]
  """
  def __init__(self):
    HTMLParser.__init__(self)
    self.flag = 0
    self.links = []
    self.title = ""
    self.img = ""
    self.content = ""

  def handle_starttag(self, tag, attrs):
    if tag == "a":
      if len(attrs) != 0:
        for (variable, value) in attrs:
          if variable == "href":
            self.links.append(value)


def get_urls_from_seed(url):
  """
  get a list of urls from a seeding url, return a list of urls

  :param str url: a full/absolute url, e.g. http://www.cnn.com/logs/
  :return: a list of full/absolute urls.
  """

  if not url or type(url) != str or not naarad.utils.is_valid_url(url):
    logger.error("get_urls_from_seed() does not have valid seeding url.")
    return

  # Extract the host info of "http://host:port/" in case of href urls are elative urls (e.g., /path/gc.log)
  # Then join (host info and relative urls) to form the complete urls
  base_index = url.find('/', len("https://"))   # get the first "/" after http://" or "https://"; handling both cases.
  base_url = url[:base_index]      # base_url = "http://host:port" or https://host:port" or http://host" (where no port is given)

  # Extract the "href" denoted urls
  urls = []
  try:
    response = urllib2.urlopen(url)
    hp = HTMLLinkExtractor()
    hp.feed(response.read())
    urls = hp.links
    hp.close()
  except urllib2.HTTPError:
    logger.error("Got HTTPError when opening the url of %s" % url)
    return urls

  # Check whether the url is relative or complete
  for i in range(len(urls)):
    if not urls[i].startswith("http://") and not urls[i].startswith("https://"):    # a relative url ?
      urls[i] = base_url + urls[i]

  return urls


def download_url_single(inputs, outdir, outfile=None):
  """
  Downloads a http(s) url to a local file
  :param str inputs:  the absolute url
  :param str outdir: Required. the local directory to put the downloadedfiles.
  :param str outfile: // Optional. If this is given, the downloaded url will be renated to outfile;
    If this is not given, then the local file will be the original one, as given in url.
  :return: the local full path name of downloaded url
  """

  if not inputs or type(inputs) != str or not outdir or type(outdir) != str:
    logging.error("The call parameters are invalid.")
    return
  else:
    if not os.path.exists(outdir):
      os.makedirs(outdir)

  output_file = handle_single_url(inputs, outdir, outfile)
  return output_file


def download_url_regex(inputs, outdir, regex=".*"):
  """
  Downloads http(s) urls to a local files
  :param str inputs: Required, the seed url
  :param str outdir: Required. the local directory to put the downloadedfiles.
  :param str regex: Optional, a regex string. If not given, then all urls will be valid
  :return: A list of local full path names (downloaded from inputs)
  """
  if not inputs or type(inputs) != str \
     or not outdir or type(outdir) != str:
    logging.error("The call parameters are invalid.")
    return
  else:
    if not os.path.exists(outdir):
      os.makedirs(outdir)

  output_files = []
  files = get_urls_from_seed(inputs)
  for f in files:
    if re.compile(regex).match(f):
      output_file = handle_single_url(f, outdir)
      output_files.append(output_file)

  return output_files


def download_url_list(url_list, outdir):
  """
  Downloads list of http(s) urls to local files
  :param list url_list: list of URLs to download
  :param str outdir: Required. the local directory to put the downloadedfiles.
    If this is not given, then the local file will be the original one, as given in url.
  :return None
  """
  for url in url_list:
    download_url_single(url, outdir)



# coding=utf-8
"""
Copyright 2013 LinkedIn Corp. All rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import sys
# add the path of ~/naarad/src;   the testing py is under ~/naarad/test
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), 'src')))

from naarad.run_steps.run_step import Run_Step
from naarad.run_steps.local_cmd import Local_Cmd
import naarad.utils
import naarad.naarad_constants as CONSTANTS

local_cmd_obj = None


def setup_module():
  global local_cmd_obj
  run_cmd = "sleep 10"
  run_rank = 1
  run_type = CONSTANTS.RUN_TYPE_WORKLOAD
  run_order = CONSTANTS.PRE_ANALYSIS_RUN
  call_type = 'local'
  local_cmd_obj = Local_Cmd(run_type, run_cmd, call_type, run_order, run_rank)


def test_run_local_cmd():
  """
  Test whether local command works as expected
  :return: None
  """
  global local_cmd_obj
  local_cmd_obj.run()
  ts_diff = naarad.utils.convert_to_unixts(local_cmd_obj.ts_end) - naarad.utils.convert_to_unixts(local_cmd_obj.ts_start)
  ts_diff /= 1000
  assert ts_diff == 10.0


def test_run_local_cmd_with_kill():
  """
  Test whether local command works as expected when kill is specified
  :return: None
  """
  global local_cmd_obj
  local_cmd_obj.kill_after_seconds = 5
  local_cmd_obj.run()
  ts_diff = naarad.utils.convert_to_unixts(local_cmd_obj.ts_end) - naarad.utils.convert_to_unixts(local_cmd_obj.ts_start)
  ts_diff /= 1000
  assert ts_diff == 5.0



# coding=utf-8
"""
Copyright 2013 LinkedIn Corp. All rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from collections import defaultdict
import datetime
import gc
import logging
import os
import re
import numpy
from naarad.metrics.metric import Metric
import naarad.utils
import sys

logger = logging.getLogger('naarad.metrics.cluster_metric')


class ClusterMetric(Metric):
  """
  supporting the metric of Cluster, which aggregates the performance metrics of multiple hosts
  """

  metrics = []   # all other non-aggregate metrics;
  aggr_metrics = []  # metrics to be aggregated
  aggr_hosts = []  # hosts to be aggregated

  def __init__(self, section, aggregate_hosts, aggregate_metrics, metrics, output_directory, resource_path, label,
               ts_start, ts_end, rule_strings, important_sub_metrics, anomaly_detection_metrics, **other_options):
    self.metrics = metrics
    self.aggr_hosts = aggregate_hosts.split()

    # Metric arguments take 'infile' and 'hostname', for ClusterMetric, they are invalid, so just provide empty strings.
    Metric.__init__(self, section, '', '', '', output_directory, resource_path, label, ts_start, ts_end, rule_strings,
                    important_sub_metrics, anomaly_detection_metrics)
    self.aggr_metrics = aggregate_metrics.split()

    for (key, val) in other_options.iteritems():
      setattr(self, key, val.split())

  def collect(self):
    """
    Take a list of metrics, filter all metrics based on hostname, and metric_type
    For each metric, merge the corresponding csv files into one,update corresponding properties such as csv_column_map.
    Users can specify functions: raw, count (qps), sum (aggregated value), avg (averaged value)
    The timestamp granularity of aggregated submetrics is in seconds (sub-second is not supported)
    """

    for aggr_metric in self.aggr_metrics:   # e.g., SAR-device.sda.await:count,sum,avg
      functions_aggr = []
      fields = aggr_metric.split(":")
      cur_metric_type = fields[0].split(".")[0]  # e.g. SAR-device

      if len(fields) > 1:  # The user has to specify the aggregate functions (i.e., :raw,count,sum,avg)
        func_user = ''.join(fields[1].split())
        functions_aggr.extend(func_user.split(","))
      else:  # no user input of aggregate functions
        return True

      cur_column = '.'.join(fields[0].split('.')[1:])    # e.g. sda.await or all.percent-sys

      # Store data points of various aggregation functions
      aggr_data = {}
      aggr_data['raw'] = []   # Store all the raw values
      aggr_data['sum'] = defaultdict(float)   # Store the sum values for each timestamp
      aggr_data['count'] = defaultdict(int)  # Store the count of each timestamp (i.e. qps)

      for metric in self.metrics:   # Loop the list to find from all metrics to merge
        if metric.hostname in self.aggr_hosts and \
           cur_column in metric.csv_column_map.values():
          file_csv = metric.get_csv(cur_column)
          timestamp_format = None
          with open(file_csv) as fh:
            for line in fh:
              aggr_data['raw'].append(line.rstrip())
              words = line.split(",")
              ts = words[0].split('.')[0]   # In case of sub-seconds; we only want the value of seconds;
              if not timestamp_format or timestamp_format == 'unknown':
                timestamp_format = naarad.utils.detect_timestamp_format(ts)
              if timestamp_format == 'unknown':
                continue
              ts = naarad.utils.get_standardized_timestamp(ts, timestamp_format)
              aggr_data['sum'][ts] += float(words[1])
              aggr_data['count'][ts] += 1
      # "raw" csv file
      if 'raw' in functions_aggr:
        out_csv = self.get_csv(cur_column, 'raw')
        self.csv_files.append(out_csv)
        with open(out_csv, 'w') as fh:
          fh.write("\n".join(sorted(aggr_data['raw'])))

      # "sum"  csv file
      if 'sum' in functions_aggr:
        out_csv = self.get_csv(cur_column, 'sum')
        self.csv_files.append(out_csv)
        with open(out_csv, 'w') as fh:
          for (k, v) in sorted(aggr_data['sum'].items()):
            fh.write(k + "," + str(v) + '\n')

      # "avg" csv file
      if 'avg' in functions_aggr:
        out_csv = self.get_csv(cur_column, 'avg')
        self.csv_files.append(out_csv)
        with open(out_csv, 'w') as fh:
          for (k, v) in sorted(aggr_data['sum'].items()):
            fh.write(k + "," + str(v / aggr_data['count'][k]) + '\n')

      # "count" csv file (qps)
      if 'count' in functions_aggr:
        out_csv = self.get_csv(cur_column, 'count')
        self.csv_files.append(out_csv)
        with open(out_csv, 'w') as fh:
          for (k, v) in sorted(aggr_data['count'].items()):
            fh.write(k + "," + str(v) + '\n')

      gc.collect()
    return True

  def get_csv(self, column, func):
    csv_file = Metric.get_csv(self, column + '.' + func)
    return csv_file

  def parse(self):
    """
    Merge multiple hosts' csv into one csv file. This approach has the benefit of reusing calculate_stats(), but with the penalty of reading the single csv
    later for calculate_stats(). However, since file cache will cache the newly written csv files, reading the csv file will not likely be a IO bottleneck.
    """

    return True



# coding=utf-8
"""
Copyright 2013 LinkedIn Corp. All rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import naarad.utils
import logging
import os

from naarad.metrics.metric import Metric

logger = logging.getLogger('naarad.metrics.INNOMetric')


class INNOMetric(Metric):
  C_MAX_COMMANDS = 10
  graph_lib = None

  def __init__(self, metric_type, infile, hostname, aggr_metrics, outdir, resource_path, label, ts_start, ts_end, rule_strings, important_sub_metrics,
               anomaly_detection_metrics, **other_options):
    Metric.__init__(self, metric_type, infile, hostname, aggr_metrics, outdir, resource_path, label, ts_start, ts_end, rule_strings, important_sub_metrics,
                    anomaly_detection_metrics)
    for (key, val) in other_options.iteritems():
      setattr(self, key, val.split())

  def get_csv_C(self, command, column):
    outcsv = os.path.join(self.resource_directory, "{0}.{1}.{2}.csv".format(self.metric_type, command, column))
    self.csv_column_map[outcsv] = command + '.' + column
    return outcsv

  def parse(self):
    logger.info("Working on innotop metric: %s", self.infile)
    if self.metric_type == "INNOTOP-C":
      return self.parse_innotop_mode_c()
    elif self.metric_type == "INNOTOP-M":
      return self.parse_innotop_mode_m()
    else:
      return self.parse_innotop_mode_b()

  def parse_innotop_mode_c(self):
    with open(self.infile, 'r') as infh:
      headerline = infh.readline()
      columns = headerline.split()[2:]
      outfilehandlers = {}
      for line in infh:
        l = line.strip().split(' ', 1)
        if len(l) <= 1:
          continue
        ts = l[0].strip().replace('T', ' ')
        try:
          nameval = l[1].strip().split('\t', 1)
        except IndexError:
          logger.warn("Badly formatted line: %s", line)
          logger.warn("Expected tab separated values")
          continue
        command = nameval[0]
        if command not in outfilehandlers:
          # Only looking at top N commands
          if len(outfilehandlers) > self.C_MAX_COMMANDS:
            continue
          # TODO(rmaheshw) : Use collections.defaultdict instead to avoid initializing dicts
          outfilehandlers[command] = {}
        words = nameval[1].split('\t')
        for i in range(len(words)):
          if self.options and columns[i] not in self.options:
            continue
          if columns[i] not in outfilehandlers[command]:
            outfilehandlers[command][columns[i]] = open(self.get_csv_C(command, columns[i]), 'w')
            self.csv_files.append(self.get_csv_C(command, columns[i]))
          ts = naarad.utils.reconcile_timezones(ts, self.timezone, self.graph_timezone)
          outfilehandlers[command][columns[i]].write(ts + ',')
          outfilehandlers[command][columns[i]].write(words[i])
          outfilehandlers[command][columns[i]].write('\n')
      for command in outfilehandlers:
        for column in outfilehandlers[command]:
          outfilehandlers[command][column].close()
    return True

  def parse_innotop_mode_b(self):
    """ Generic parsing method for all other modes """
    with open(self.infile, 'r') as infh:
      # Pre processing to figure out different headers
      max_row_quot = 0
      valrow = -1
      thisrowcolumns = {}
      data = {}
      while True:
        line1 = infh.readline()
        words = line1.split()
        # special case for -I (iostat) option
        # skipping all the 'thread' lines
        if words[1] == "thread" and self.metric_type == "INNOTOP-I":
          while True:
            line1 = infh.readline()
            words = line1.split()
            if naarad.utils.is_number(words[1]):
              line1 = infh.readline()
            else:
              break
        if words[1] == "thread" and self.metric_type == "INNOTOP-R":
          break
        # Skip next line
        infh.readline()
        last_ts = words[0].strip().replace('T', ' ')
        if not naarad.utils.is_number(words[1]):
          thisrowcolumns[max_row_quot] = words[1:]
          for column in words[1:]:
            if self.options and column not in self.options:
              continue
            data[column] = []
          if self.metric_type == "INNOTOP-I":
            data["check_pt_age"] = []
          max_row_quot += 1
        else:
          break
      # infh.seek(0)
      # Real Processing
      for line in infh:
        l = line.strip().split(' ', 1)
        if len(l) <= 1:
          continue
        ts = l[0].strip().replace('T', ' ')
        if not ts == last_ts:
          last_ts = ts
          valrow = -1
        try:
          words = l[1].strip().split('\t')
        except IndexError:
          logger.warn("Bad line: %s", line)
          continue
        # special case for -I (iostat) option
        # skipping all the 'thread' lines
        if words[0] == "thread" or (naarad.utils.is_number(words[0]) and "thread" in words[1]):
          continue
        if naarad.utils.is_number(words[0]):
          valrow += 1
          quot = valrow % max_row_quot
          # Special case for -R, skipping all 'thread' value lines
          if quot >= len(thisrowcolumns):
            continue
          columns = thisrowcolumns[quot]
          if len(words) > len(columns):
            continue
          for i in range(len(words)):
            if self.options and columns[i] not in self.options:
              continue
            column = columns[i]
            # Converting -- to 0, seen this for buf_pool_hit_rate
            if words[i] == "--":
              words[i] = "0"
            ts = naarad.utils.reconcile_timezones(ts, self.timezone, self.graph_timezone)
            # Calculating check point age
            if self.metric_type == "INNOTOP-I":
              if column == "log_seq_no":
                log_seq_no = int(words[i])
              elif column == "log_flushed_to":
                check_pt_age = log_seq_no - int(words[i])
                tup = [ts, str(check_pt_age)]
                data["check_pt_age"].append(tup)
            tup = [ts, words[i]]
            data[column].append(tup)
    # Post Proc, writing the different out files
    for column in data:
      csvfile = self.get_csv(column)
      self.csv_files.append(csvfile)
      with open(csvfile, 'w') as outfh:
        for tup in data[column]:
          outfh.write(','.join(tup))
          outfh.write('\n')
    return True

  def parse_innotop_mode_m(self):
    """ Special parsing method for Innotop "Replication Status" results (innotop --mode M)"""
    with open(self.infile, 'r') as infh:
      # Pre processing to figure out different headers
      max_row_quot = 0
      valrow = -1
      thisrowcolumns = {}
      data = {}
      last_ts = None
      while True:
        # 2012-05-11T00:00:02 master_host slave_sql_running  time_behind_master  slave_catchup_rate  slave_open_temp_tables  relay_log_pos   last_error
        line1 = infh.readline()
        words = line1.split()
        # Skip next line
        infh.readline()
        is_header = True
        for word in words:
          if naarad.utils.is_number(word):
            last_ts = words[0].strip().replace('T', ' ')
            is_header = False
            break  # from this loop
        if len(words) > 2 and is_header:
          thisrowcolumns[max_row_quot] = words[2:]
          for column in thisrowcolumns[max_row_quot]:
            data[column] = []
          max_row_quot += 1
        else:
          break
          # from pre-processing. All headers accounted for

      # Real Processing
      if not last_ts:
        logger.warn("last_ts not set, looks like there is no data in file %s", self.infile)
        return True
      infh.seek(0)
      is_bad_line = False
      outfilehandlers = {}
      for line in infh:
        l = line.strip().split(' ', 1)
        # Blank line
        if len(l) <= 1:
          continue
        ts = l[0].strip().replace('T', ' ')
        if ts != last_ts:
          last_ts = ts
          valrow = -1
        nameval = l[1].strip().split('\t', 1)
        try:
          words = nameval[1].split('\t')
        except IndexError:
          logger.warn("Bad line: %s", line)
          continue
        valrow += 1
        command = nameval[0]
        if command not in outfilehandlers:
          outfilehandlers[command] = {}
        quot = valrow % max_row_quot
        columns = thisrowcolumns[quot]
        for i in range(len(words)):
          if len(words) > len(columns):
            logger.warn("Mismatched number of columns: %s", line)
            logger.warn("%d %d", len(words), len(columns))
            break
          if words[i] in columns:
            logger.warn("Skipping line: %s", line)
            valrow -= 1
            break
          if self.options and columns[i] not in self.options:
            continue
          if columns[i] not in outfilehandlers[command]:
            outfilehandlers[command][columns[i]] = open(self.get_csv_C(command, columns[i]), 'w')
            self.csv_files.append(self.get_csv_C(command, columns[i]))
          ts = naarad.utils.reconcile_timezones(ts, self.timezone, self.graph_timezone)
          outfilehandlers[command][columns[i]].write(ts + ',')
          outfilehandlers[command][columns[i]].write(words[i])
          outfilehandlers[command][columns[i]].write('\n')
      for command in outfilehandlers:
        for column in outfilehandlers[command]:
          outfilehandlers[command][column].close()
    return True



# coding=utf-8
"""
Copyright 2013 LinkedIn Corp. All rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import datetime
import logging
import os
import re
import sys
import threading

from naarad.metrics.metric import Metric
import naarad.utils
from naarad.naarad_constants import important_sub_metrics_import

logger = logging.getLogger('naarad.metrics.GCMetric')


class GCMetric(Metric):
  """ Class for GC logs, deriving from class Metric """

  bin_path = os.path.dirname(sys.argv[0])
  clock_format = '%Y-%m-%d %H:%M:%S'
  rate_types = ()
  val_types = ('alloc', 'promo', 'used0', 'used1', 'used', 'commit0', 'commit1', 'commit', 'gen0', 'gen0t', 'gen0usr', 'gen0sys', 'gen0real',
      'cmsIM', 'cmsRM', 'cmsRS', 'GCPause', 'cmsCM', 'cmsCP', 'cmsCS', 'cmsCR', 'safept', 'apptime', 'used0AfterGC', 'used1AfterGC', 'usedAfterGC',
      'gen1t', 'g1-pause-young', 'g1-pause-mixed', 'g1-pause-remark', 'g1-pause-cleanup', 'g1-pause-remark.ref-proc', 'g1-pause-young.parallel',
      'g1-pause-young.parallel.gcworkers', 'g1-pause-young.parallel.ext-root-scanning.avg', 'g1-pause-young.parallel.ext-root-scanning.max',
      'g1-pause-young.parallel.update-rs.avg', 'g1-pause-young.parallel.update-rs.max', 'g1-pause-young.parallel.update-rs.processed-buffers.avg',
      'g1-pause-young.parallel.update-rs.processed-buffers.max', 'g1-pause-young.parallel.scan-rs.avg', 'g1-pause-young.parallel.scan-rs.max',
      'g1-pause-young.parallel.object-copy-rs.avg', 'g1-pause-young.parallel.object-copy-rs.max', 'g1-pause-young.parallel.termination.avg',
      'g1-pause-young.parallel.termination.max', 'g1-pause-young.parallel.gc-worker-other.avg', 'g1-pause-young.parallel.gc-worker-other.max',
      'g1-pause-young.parallel.gc-worker-total.avg', 'g1-pause-young.parallel.gc-worker-total.max', 'g1-pause-young.parallel.gc-worker-end.avg',
      'g1-pause-young.parallel.gc-worker-end.max', 'g1-pause-young.code-root-fixup', 'g1-pause-young.clear-ct', 'g1-pause-young.other',
      'g1-pause-young.other.choose-cset', 'g1-pause-young.other.ref-proc', 'g1-pause-young.other.reg-enq', 'g1-pause-young.other.free-cset',
      'g1-pause-mixed.parallel', 'g1-pause-mixed.parallel.gcworkers', 'g1-pause-mixed.parallel.ext-root-scanning.avg',
      'g1-pause-mixed.parallel.ext-root-scanning.max', 'g1-pause-mixed.parallel.update-rs.avg', 'g1-pause-mixed.parallel.update-rs.max',
      'g1-pause-mixed.parallel.update-rs.processed-buffers.avg', 'g1-pause-mixed.parallel.update-rs.processed-buffers.max',
      'g1-pause-mixed.parallel.scan-rs.avg', 'g1-pause-mixed.parallel.scan-rs.max', 'g1-pause-mixed.parallel.object-copy-rs.avg',
      'g1-pause-mixed.parallel.object-copy-rs.max', 'g1-pause-mixed.parallel.termination.avg', 'g1-pause-mixed.parallel.termination.max',
      'g1-pause-mixed.parallel.gc-worker-other.avg', 'g1-pause-mixed.parallel.gc-worker-other.max', 'g1-pause-mixed.parallel.gc-worker-total.avg',
      'g1-pause-mixed.parallel.gc-worker-total.max', 'g1-pause-mixed.parallel.gc-worker-end.avg', 'g1-pause-mixed.parallel.gc-worker-end.max',
      'g1-pause-mixed.code-root-fixup', 'g1-pause-mixed.clear-ct', 'g1-pause-mixed.other', 'g1-pause-mixed.other.choose-cset',
      'g1-pause-mixed.other.ref-proc', 'g1-pause-mixed.other.reg-enq', 'g1-pause-mixed.other.free-cset', 'g1-pause-young.parallel.gc-worker-start.avg',
      'g1-pause-young.parallel.gc-worker-start.max', 'g1-pause-mixed.parallel.gc-worker-start.avg', 'g1-pause-mixed.parallel.gc-worker-start.max',
      'g1-eden-occupancy-before-gc', 'g1-eden-capacity-before-gc', 'g1-eden-occupancy-after-gc', 'g1-eden-capacity-after-gc', 'g1-survivor-before-gc',
      'g1-survivor-after-gc', 'g1-heap-occupancy-before-gc', 'g1-heap-capacity-before-gc', 'g1-heap-occupancy-after-gc', 'g1-heap-capacity-after-gc',
      'g1-young-cpu.sys', 'g1-young-cpu.usr', 'g1-young-cpu.real', 'g1-mixed-cpu.usr', 'g1-mixed-cpu.sys', 'g1-mixed-cpu.real')

  def __init__(self, metric_type, infile_list, hostname, aggr_metrics, outdir, resource_path, label, ts_start, ts_end, rule_strings,
               important_sub_metrics, anomaly_detection_metrics, **other_options):
    Metric.__init__(self, metric_type, infile_list, hostname, aggr_metrics, outdir, resource_path, label, ts_start, ts_end, rule_strings,
                    important_sub_metrics, anomaly_detection_metrics)
    if not self.important_sub_metrics:
      self.important_sub_metrics = important_sub_metrics_import['GC']
    self.sub_metrics = self.val_types
    self.beginning_ts = None
    self.beginning_date = None
    for (key, val) in other_options.iteritems():
      if key == 'gc-options' or key == 'sub_metrics':
        self.sub_metrics = val.split()
      else:
        setattr(self, key, val)
    self.sub_metric_description = {
        'gen0': 'young gen collection time, excluding gc_prologue & gc_epilogue',
        'gen0t': 'young gen collection time, including gc_prologue & gc_epilogue',
        'gen0usr': 'young gen collection time in cpu user secs',
        'gen0sys': 'young gen collection time in cpu sys secs',
        'gen0real': 'young gen collection time in elapsed secs',
        'gen1i': 'train generation incremental collection',
        'gen1t': 'old generation collection or full GC',
        'cmsIM': 'CMS initial mark pause',
        'cmsRM': 'CMS remark pause',
        'cmsRS': 'CMS resize pause',
        'GCPause': 'all stop-the-world GC pauses',
        'cmsCM': 'CMS concurrent mark phase',
        'cmsCP': 'CMS concurrent preclean phase',
        'cmsCS': 'CMS concurrent sweep phase',
        'cmsCR': 'CMS concurrent reset phase',
        'alloc': 'object allocation in MB (approximate***)',
        'promo': 'object promotion in MB (approximate***)',
        'used0': 'young gen used memory size (before gc)',
        'used1': 'old gen used memory size (before gc)',
        'used': 'heap space used memory size (before gc) (excludes perm gen)',
        'commit0': 'young gen committed memory size (after gc)',
        'commit1': 'old gen committed memory size (after gc)',
        'commit': 'heap committed memory size (after gc) (excludes perm gen)',
        'apptime': 'amount of time application threads were running',
        'safept': 'amount of time the VM spent at safepoints (app threads stopped)',
        'used0AfterGC': 'young gen used memory size (after gc)',
        'used1AfterGC': 'old gen used memory size (after gc)',
        'usedAfterGC': 'heap space used memory size (after gc)',
        'g1-pause-young': 'G1 Young GC Pause (seconds)',
        'g1-pause-mixed': 'G1 Mixed GC Pause (seconds)',
        'g1-pause-remark': 'G1 Remark Pause (seconds)',
        'g1-pause-cleanup': 'G1 Cleanup Pause (seconds)',
        'g1-pause-remark.ref-proc': 'G1 Remark: Reference Processing (seconds)',
        'g1-pause-young.parallel': 'G1 Young GC Pause: Parallel Operations (ms)',
        'g1-pause-young.parallel.gcworkers': 'G1 Young GC Pause: Number of Parallel GC Workers',
        'g1-pause-young.parallel.gc-worker-start.avg': 'G1 Young GC Pause : Parallel : Avg Time spent in GC worker start (ms)',
        'g1-pause-young.parallel.gc-worker-start.max': 'G1 Young GC Pause : Parallel : Max Time spent in GC worker start (ms)',
        'g1-pause-young.parallel.ext-root-scanning.avg': 'G1 Young GC Pause: Avg Time spent in ext-root-scanning',
        'g1-pause-young.parallel.ext-root-scanning.max': 'G1 Young GC Pause: Max Time spent in ext-root-scanning',
        'g1-pause-young.parallel.update-rs.avg': 'G1 Young GC Pause: Parallel : Avg Time spent in updating Rsets',
        'g1-pause-young.parallel.update-rs.max': 'G1 Young GC Pause: Parallel : Max Time spent in updating Rsets',
        'g1-pause-young.parallel.update-rs.processed-buffers.avg': 'G1 Young GC Pause : Parallel : Update Rset : Avg number of processed buffers',
        'g1-pause-young.parallel.update-rs.processed-buffers.max': 'G1 Young GC Pause : Parallel : Update Rset : Max number of processed buffers',
        'g1-pause-young.parallel.scan-rs.avg': 'G1 Young GC Pause: Parallel : Avg Time spent in scanning Rsets',
        'g1-pause-young.parallel.scan-rs.max': 'G1 Young GC Pause: Parallel : Max Time spent in scannning Rsets',
        'g1-pause-young.parallel.object-copy-rs.avg': 'G1 Young GC Pause : Parallel : Avg Time spent in Object Copy',
        'g1-pause-young.parallel.object-copy-rs.max': 'G1 Young GC Pause : Parallel : Max Time spent in Object Copy',
        'g1-pause-young.parallel.termination.avg': 'G1 Young GC Pause : Parallel : Avg Time spent in termination',
        'g1-pause-young.parallel.termination.max': 'G1 Young GC Pause : Parallel : Max Time spent in termination',
        'g1-pause-young.parallel.gc-worker-other.avg': 'G1 Young GC Pause : Parallel : Avg Time spent in other',
        'g1-pause-young.parallel.gc-worker-other.max': 'G1 Young GC Pause : Parallel : Max Time spent in other',
        'g1-pause-young.parallel.gc-worker-total.avg': 'G1 Young GC Pause : Parallel : Avg Total time for GC worker',
        'g1-pause-young.parallel.gc-worker-total.max': 'G1 Young GC Pause : Parallel : Max Total time for GC worker',
        'g1-pause-young.parallel.gc-worker-end.avg': 'G1 Young GC Pause : Parallel : Avg Time for GC worker end',
        'g1-pause-young.parallel.gc-worker-end.max': 'G1 Young GC Pause : Parallel : Max Time for GC worker end',
        'g1-pause-young.code-root-fixup': 'G1 Young GC Pause : Time spent in code root fixup (ms)',
        'g1-pause-young.clear-ct': 'G1 Young GC Pause: Time spent in clear ct (ms)',
        'g1-pause-young.other': 'G1 Young GC Pause: Time spent in other (ms)',
        'g1-pause-young.other.choose-cset': 'G1 Young GC Pause : Other : Time spent in choosing CSet (ms)',
        'g1-pause-young.other.ref-proc': 'G1 Young GC Pause : Other : Time spent in reference processing (ms)',
        'g1-pause-young.other.reg-enq': 'G1 Young GC Pause : Other : Time spent in reg-enq(ms)',
        'g1-pause-young.other.free-cset': 'G1 Young GC Pause : Other : Time spent in processing free Cset(ms)',
        'g1-pause-mixed.parallel': 'G1 Mixed GC Pause: Parallel Operations (ms)',
        'g1-pause-mixed.parallel.gcworkers': 'G1 Mixed GC Pause: Number of Parallel GC Workers',
        'g1-pause-mixed.parallel.gc-worker-start.avg': 'G1 Mixed GC Pause : Parallel : Avg Time spent in GC worker start (ms)',
        'g1-pause-mixed.parallel.gc-worker-start.max': 'G1 Mixed GC Pause : Parallel : Max Time spent in GC worker start (ms)',
        'g1-pause-mixed.parallel.ext-root-scanning.avg': 'G1 Mixed GC Pause: Avg Time spent in ext-root-scanning',
        'g1-pause-mixed.parallel.ext-root-scanning.max': 'G1 Mixed GC Pause: Max Time spent in ext-root-scanning',
        'g1-pause-mixed.parallel.update-rs.avg': 'G1 Mixed GC Pause: Parallel : Avg Time spent in updating Rsets',
        'g1-pause-mixed.parallel.update-rs.max': 'G1 Mixed GC Pause: Parallel : Max Time spent in updating Rsets',
        'g1-pause-mixed.parallel.update-rs.processed-buffers.avg': 'G1 Mixed GC Pause : Parallel : Update Rset : Avg number of processed buffers',
        'g1-pause-mixed.parallel.update-rs.processed-buffers.max': 'G1 Mixed GC Pause : Parallel : Update Rset : Max number of processed buffers',
        'g1-pause-mixed.parallel.scan-rs.avg': 'G1 Mixed GC Pause: Parallel : Avg Time spent in scanning Rsets',
        'g1-pause-mixed.parallel.scan-rs.max': 'G1 Mixed GC Pause: Parallel : Max Time spent in scannning Rsets',
        'g1-pause-mixed.parallel.object-copy-rs.avg': 'G1 Mixed GC Pause : Parallel : Avg Time spent in Object Copy',
        'g1-pause-mixed.parallel.object-copy-rs.max': 'G1 Mixed GC Pause : Parallel : Max Time spent in Object Copy',
        'g1-pause-mixed.parallel.termination.avg': 'G1 Mixed GC Pause : Parallel : Avg Time spent in termination',
        'g1-pause-mixed.parallel.termination.max': 'G1 Mixed GC Pause : Parallel : Max Time spent in termination',
        'g1-pause-mixed.parallel.gc-worker-other.avg': 'G1 Mixed GC Pause : Parallel : Avg Time spent in other',
        'g1-pause-mixed.parallel.gc-worker-other.max': 'G1 Mixed GC Pause : Parallel : Max Time spent in other',
        'g1-pause-mixed.parallel.gc-worker-total.avg': 'G1 Mixed GC Pause : Parallel : Avg Total time for GC worker',
        'g1-pause-mixed.parallel.gc-worker-total.max': 'G1 Mixed GC Pause : Parallel : Max Total time for GC worker',
        'g1-pause-mixed.parallel.gc-worker-end.avg': 'G1 Mixed GC Pause : Parallel : Avg Time for GC worker end',
        'g1-pause-mixed.parallel.gc-worker-end.max': 'G1 Mixed GC Pause : Parallel : Max Time for GC worker end',
        'g1-pause-mixed.code-root-fixup': 'G1 Mixed GC Pause : Time spent in code root fixup (ms)',
        'g1-pause-mixed.clear-ct': 'G1 Mixed GC Pause: Time spent in clear ct (ms)',
        'g1-pause-mixed.other': 'G1 Mixed GC Pause: Time spent in other (ms)',
        'g1-pause-mixed.other.choose-cset': 'G1 Mixed GC Pause : Other : Time spent in choosing CSet (ms)',
        'g1-pause-mixed.other.ref-proc': 'G1 Mixed GC Pause : Other : Time spent in reference processing (ms)',
        'g1-pause-mixed.other.reg-enq': 'G1 Mixed GC Pause : Other : Time spent in reg-enq(ms)',
        'g1-pause-mixed.other.free-cset': 'G1 Mixed GC Pause : Other : Time spent in processing free Cset(ms)',
        'g1-eden-occupancy-before-gc': 'G1 Eden Occupancy (MB) (Before GC)',
        'g1-eden-capacity-before-gc': 'G1 Eden Capacity (MB) (Before GC)',
        'g1-eden-occupancy-after-gc': 'G1 Eden Occupancy (MB) (After GC)',
        'g1-eden-capacity-after-gc': 'G1 Eden Capacity (MB) (After GC)',
        'g1-survivor-before-gc': 'G1 Survivor Size (MB) (Before GC)',
        'g1-survivor-after-gc': 'G1 Survivor Size (MB) (After GC)',
        'g1-heap-occupancy-before-gc': 'G1 Heap Occupancy (MB) (Before GC)',
        'g1-heap-capacity-before-gc': 'G1 Heap Capacity (MB) (Before GC)',
        'g1-heap-occupancy-after-gc': 'G1 Heap Occupancy (MB) (After GC)',
        'g1-heap-capacity-after-gc': 'G1 Heap Capacity (MB) (After GC)',
        'g1-young-cpu.sys': 'G1 Young GC : sys cpu time (seconds)',
        'g1-young-cpu.usr': 'G1 Young GC : usr cpu time (seconds)',
        'g1-young-cpu.real': 'G1 Young GC : elapsed time (seconds)',
        'g1-mixed-cpu.usr': 'G1 Mixed GC : usr cpu time (seconds)',
        'g1-mixed-cpu.sys': 'G1 Mixed GC : sys cpu time (seconds)',
        'g1-mixed-cpu.real': 'G1 Mixed GC : elapsed time (seconds)'
    }

  def parse(self):
    prefix = os.path.join(self.resource_directory, self.label)
    awk_cmd = os.path.join(self.bin_path, 'PrintGCStats')
    gc_metrics = set(self.val_types) & set(self.sub_metrics)
    if self.ts_start:
      awk_cmd += ' -v ts_start="' + naarad.utils.get_standardized_timestamp(self.ts_start, None) + '"'
    if self.ts_end:
      awk_cmd += ' -v ts_end="' + naarad.utils.get_standardized_timestamp(self.ts_end, None) + '"'
    cmd = "{0} -v plot={1} -v splitfiles=1 -v datestamps=1 -v plotcolumns=2 -v splitfileprefix={2} {3}".format(awk_cmd, ','.join(gc_metrics), prefix,
                                                                                                               ' '.join(self.infile_list))
    logger.info("Parsing GC metric with cmd: %s", cmd)
    os.system(cmd)
    for gc_sub_metric in gc_metrics:
      outcsv = self.get_csv(gc_sub_metric)
      if naarad.utils.is_valid_file(outcsv):
        self.csv_files.append(outcsv)
    return True



# coding=utf-8
"""
Copyright 2013 LinkedIn Corp. All rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import datetime
import pygal
import os
import logging

logger = logging.getLogger('naarad.graphing.pygal_naarad')


def convert_to_date(date_str):
  return datetime.datetime.strptime(date_str, '%Y-%m-%d %H:%M:%S')


def curate_plot_list(plots):
  delete_nodes = []
  for plot in plots:
    if os.path.exists(plot.input_csv):
      if not os.path.getsize(plot.input_csv):
        logger.warning("%s file is empty. No plot corresponding to this file will be generated", plot.input_csv)
        delete_nodes.append(plot)
    else:
      logger.warning("%s file does not exist. No plot corresponding to this file will be generated", plot.input_csv)
      delete_nodes.append(plot)
  for node in delete_nodes:
    plots.remove(node)
  return plots


def graph_data(list_of_plots, output_directory, resource_path, output_filename):
  date_plot = pygal.DateY(x_label_rotation=20, height=500, width=1200, legend_at_bottom=True, style=pygal.style.BlueStyle)
  for plot in list_of_plots:
    plot_data = []
    with open(plot.input_csv, 'r') as csv_data:
      for line in csv_data:
        line_data = line.strip('\n').split(',')
        if '.' in line_data[0]:
          plot_data.append((datetime.datetime.strptime(line_data[0], '%Y-%m-%d %H:%M:%S.%f'), float(line_data[1])))
        else:
          plot_data.append((datetime.datetime.strptime(line_data[0], '%Y-%m-%d %H:%M:%S'), float(line_data[1])))
    date_plot.add(plot.graph_title, plot_data)
    date_plot.render_to_file(os.path.join(output_directory, output_filename + '.svg'))
    with open(os.path.join(output_directory, output_filename + '.div'), 'w') as div_file:
      div_file.write('<figure><embed type="image/svg+xml" src="' + resource_path + '/' + output_filename + '.svg' + '"/></figure>')
    return True, os.path.join(output_directory, output_filename + '.div')


def graph_data_on_the_same_graph(list_of_plots, output_directory, resource_path, output_filename):
  """
  graph_data_on_the_same_graph: put a list of plots on the same graph: currently it supports CDF
  """
  logger.warning('graph_data_on_the_same_graph is currently not supported in pygal')
  return False, None



# coding=utf-8
"""
Copyright 2013 LinkedIn Corp. All rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from collections import defaultdict
import datetime
import gc
import logging
import os
import re
import numpy
from naarad.metrics.metric import Metric
import naarad.utils

logger = logging.getLogger('naarad.metrics.ProcZoneinfoMetric')


class ProcZoneinfoMetric(Metric):
  """
  logs of /proc/vmstat
  The raw log file is assumed to have a timestamp prefix of all lines. E.g. in the format of "2013-01-02 03:55:22.13456 compact_fail 36"
  The log lines can be generated by   'cat /proc/vmstat | sed "s/^/$(date +%Y-%m-%d\ %H:%M:%S.%05N)\t/" '
  """

  skipped_sub_metrics = ('protection:', 'pagesets', 'cpu:', 'count:', 'high:', 'batch:', 'vm', 'all_unreclaimable:', 'prev_priority:', 'start_pfn:',
                         'inactive_ratio:')
  processed_sub_metrics = ('min', 'high', 'scanned', 'spanned', 'present')

  zones = None   # Users can specify which zones to process/plot, e.g. zones= Node.0.zone.DMA

  def __init__(self, metric_type, infile_list, hostname, aggr_metrics, output_directory, resource_path, label, ts_start, ts_end,
               rule_strings, important_sub_metrics, anomaly_detection_metrics, **other_options):
    Metric.__init__(self, metric_type, infile_list, hostname, aggr_metrics, output_directory, resource_path, label, ts_start, ts_end,
                    rule_strings, important_sub_metrics, anomaly_detection_metrics)

    self.sub_metrics = None
    # in particular, Section can specify a subset of all metrics: sub_metrics=pages.min nr_free_pages

    for (key, val) in other_options.iteritems():
      setattr(self, key, val.split())

    self.sub_metric_description = {
        'nr_free_pages': 'Number of free pages',
        'nr_inactive_anon': 'Number of inactive anonymous pages',
        'nr_active_anon': 'Number of active anonymous pages',
        'nr_inactive_file': 'Number of inactive file cache pages',
        'nr_active_file': 'Number of active file cache pages',
    }

  def parse(self):
    """
    Parse the vmstat file
    :return: status of the metric parse
    """
    file_status = True
    for input_file in self.infile_list:
      file_status = file_status and naarad.utils.is_valid_file(input_file)
      if not file_status:
        return False

    status = True
    cur_zone = None
    cur_submetric = None
    cur_value = None
    data = {}  # stores the data of each column
    for input_file in self.infile_list:
      logger.info('Processing : %s', input_file)
      timestamp_format = None
      with open(input_file) as fh:
        for line in fh:
          words = line.replace(',', ' ').split()           # [0] is day; [1] is seconds; [2...] is field names:;
          if len(words) < 3:
            continue
          ts = words[0] + " " + words[1]
          if not timestamp_format or timestamp_format == 'unknown':
            timestamp_format = naarad.utils.detect_timestamp_format(ts)
          if timestamp_format == 'unknown':
            continue
          ts = naarad.utils.get_standardized_timestamp(ts, timestamp_format)
          if self.ts_out_of_range(ts):
            continue
          if words[2] == 'Node':  # Node 0 zone      DMA
            cols = words[2:]
            cur_zone = '.'.join(cols)
            continue
          elif words[2] == 'pages':  # pages free     3936
            cur_submetric = words[2] + '.' + words[3]  # pages.free
            cur_value = words[4]
          elif words[2] in self.processed_sub_metrics:
            cur_submetric = 'pages' + '.' + words[2]  # pages.min
            cur_value = words[3]
          elif words[2] in self.skipped_sub_metrics:
            continue
          else:   # other useful submetrics
            cur_submetric = words[2]
            cur_value = words[3]
          col = cur_zone + '.' + cur_submetric  # prefix with 'Node.0.zone.DMA.
          # only process zones specified in config
          if cur_zone and self.zones and cur_zone not in self.zones:
            continue
          self.sub_metric_unit[col] = 'pages'  # The unit of the sub metric. For /proc/zoneinfo, they are all in pages
          # only process sub_metrics specified in config.
          if self.sub_metrics and cur_submetric and cur_submetric not in self.sub_metrics:
            continue
          if col in self.column_csv_map:
            out_csv = self.column_csv_map[col]
          else:
            out_csv = self.get_csv(col)   # column_csv_map[] is assigned in get_csv()
            data[out_csv] = []
          data[out_csv].append(ts + "," + cur_value)
    # post processing, putting data in csv files;
    for csv in data.keys():
      self.csv_files.append(csv)
      with open(csv, 'w') as fh:
        fh.write('\n'.join(sorted(data[csv])))
    return status



# coding=utf-8
"""
Copyright 2013 LinkedIn Corp. All rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from collections import defaultdict
import gc
import logging
import os
from xml.etree import ElementTree
import numpy
import heapq
from naarad.metrics.metric import Metric
from naarad.graphing.plot_data import PlotData as PD
import naarad.utils
import naarad.naarad_imports
from naarad.naarad_constants import important_sub_metrics_import


logger = logging.getLogger('naarad.metrics.JmeterMetric')


class JmeterMetric(Metric):

  def __init__(self, metric_type, infile_list, hostname, aggr_metrics, output_directory, resource_path, label, ts_start, ts_end,
               rule_strings, important_sub_metrics, anomaly_detection_metrics, **other_options):
    Metric.__init__(self, metric_type, infile_list, hostname, aggr_metrics, output_directory, resource_path, label, ts_start, ts_end,
                    rule_strings, important_sub_metrics, anomaly_detection_metrics)
    self.sub_metric_description = {
        'lb': 'Transaction Name',
        'lt': 'Time to First byte',
        'ts': 'Timestamp',
        'tn': 'Transaction Name (Parent)',
        's': 'Status',
        'ResponseTime': 'Response Time',
        'rc': 'Response Code',
        'rm': 'Response Message',
        'dt': 'Data Type',
        'ResponseSize': 'Response Size',
        'qps': 'Successful Transactions per second',
        'ErrorsPerSecond': 'Errors per second',
        'DataThroughput': 'Data Throughput'
    }
    self.sub_metric_units = {
        'lt': 'ms',
        'ResponseTime': 'ms',
        'ResponseSize': 'bytes',
        'qps': 'qps',
        'DataThroughput': 'mbps',
        'ErrorsPerSecond': 'qps'
    }
    self.calculated_stats = {}
    self.aggregation_granularity = 'second'
    self.calculated_percentiles = {}
    self.summary_stats = defaultdict(dict)
    self.summary_html_content_enabled = True
    self.summary_charts = [self.label + '.Overall_Summary.div']
    if not self.important_sub_metrics:
      self.important_sub_metrics = important_sub_metrics_import['JMETER']
    if other_options:
      for (key, val) in other_options.iteritems():
        setattr(self, key, val)

  def get_csv(self, transaction_name, column):
    col = naarad.utils.sanitize_string(column)
    if col == 't':
      col = 'ResponseTime'
    elif col == 'by':
      col = 'ResponseSize'
    elif col == 'thr':
      col = 'DataThroughput'
    elif col == 'eqps':
      col = 'ErrorsPerSecond'

    if transaction_name == '__overall_summary__':
      transaction_name = 'Overall_Summary'
    csv = os.path.join(self.resource_directory, self.label + '.' + transaction_name + '.' + col + '.csv')
    self.csv_column_map[csv] = transaction_name + '.' + col
    return csv

  def aggregate_count_over_time(self, metric_store, line_data, transaction_list, aggregate_timestamp):
    """
    Organize and store the count of data from the log line into the metric store by metric type, transaction, timestamp

    :param dict metric_store: The metric store used to store all the parsed jmeter log data
    :param dict line_data: dict with the extracted k:v from the log line
    :param list transaction_list: list of transaction to be used for storing the metrics from given line
    :param string aggregate_timestamp: timestamp used for storing the raw data. This accounts for aggregation time period
    :return: None
    """
    for transaction in transaction_list:
      if line_data.get('s') == 'true':
        all_qps = metric_store['qps']
      else:
        all_qps = metric_store['eqps']
      qps = all_qps[transaction]
      if aggregate_timestamp in qps:
        qps[aggregate_timestamp] += 1
      else:
        qps[aggregate_timestamp] = 1
    return None

  def aggregate_values_over_time(self, metric_store, line_data, transaction_list, metric_list, aggregate_timestamp):
    """
    Organize and store the data from the log line into the metric store by metric type, transaction, timestamp

    :param dict metric_store: The metric store used to store all the parsed jmeter log data
    :param dict line_data: dict with the extracted k:v from the log line
    :param list transaction_list: list of transaction to be used for storing the metrics from given line
    :param list metric_list: list of metrics to extract from the log line
    :param string aggregate_timestamp: timestamp used for storing the raw data. This accounts for aggregation time period
    :return: None
    """
    for metric in metric_list:
      for transaction in transaction_list:
        metric_data = reduce(defaultdict.__getitem__, [metric, transaction, aggregate_timestamp], metric_store)
        metric_data.append(float(line_data.get(metric)))
    return None

  def average_values_for_plot(self, metric_store, data, averaging_factor):
    """
    Create the time series for the various metrics, averaged over the aggregation period being used for plots

    :param dict metric_store: The metric store used to store all the parsed jmeter log data
    :param dict data: Dict with all the metric data to be output to csv
    :param float averaging_factor: averaging factor to be used for calculating the average per second metrics
    :return: None
    """
    for metric, transaction_store in metric_store.items():
      for transaction, time_store in transaction_store.items():
        for time_stamp, metric_data in sorted(time_store.items()):
          if metric in ['t', 'by']:
            data[self.get_csv(transaction, metric)].append(','.join([str(time_stamp), str(sum(map(float, metric_data)) / float(len(metric_data)))]))
            if metric == 'by':
              metric_store['thr'][transaction][time_stamp] = sum(map(float, metric_data)) / float(averaging_factor * 1024 * 1024 / 8.0)
              data[self.get_csv(transaction, 'thr')].append(','.join([str(time_stamp), str(metric_store['thr'][transaction][time_stamp])]))
          elif metric in ['qps', 'eqps']:
            data[self.get_csv(transaction, metric)].append(','.join([str(time_stamp), str(metric_data / float(averaging_factor))]))
    return None

  def calculate_key_stats(self, metric_store):
    """
    Calculate key statistics for given data and store in the class variables calculated_stats and calculated_percentiles
    calculated_stats:
      'mean', 'std', 'median', 'min', 'max'
    calculated_percentiles:
      range(5,101,5), 99
    :param dict metric_store: The metric store used to store all the parsed jmeter log data
    :return: none
    """
    stats_to_calculate = ['mean', 'std', 'median', 'min', 'max']  # TODO: get input from user
    percentiles_to_calculate = range(5, 101, 5)  # TODO: get input from user
    percentiles_to_calculate.append(99)
    for transaction in metric_store['t'].keys():
      transaction_key = transaction + '.' + 'ResponseTime'
      # For ResponseTime and ResponseSize, each timestamp has a list of values associated with it.
      # Using heapq.merge to merge all the lists into a single list to be passed to numpy.
      self.calculated_stats[transaction_key], self.calculated_percentiles[transaction_key] = \
          naarad.utils.calculate_stats(list(heapq.merge(*metric_store['t'][transaction].values())),
                                   stats_to_calculate, percentiles_to_calculate)
      self.update_summary_stats(transaction_key)
      transaction_key = transaction + '.' + 'qps'
      if len(metric_store['qps'][transaction].values()) > 0:
        self.calculated_stats[transaction_key], self.calculated_percentiles[transaction_key] = \
            naarad.utils.calculate_stats(metric_store['qps'][transaction].values(),
                                         stats_to_calculate, percentiles_to_calculate)
        self.update_summary_stats(transaction_key)
      transaction_key = transaction + '.' + 'ResponseSize'
      self.calculated_stats[transaction_key], self.calculated_percentiles[transaction_key] = \
          naarad.utils.calculate_stats(list(heapq.merge(*metric_store['by'][transaction].values())),
                                       stats_to_calculate, percentiles_to_calculate)
      self.update_summary_stats(transaction_key)
      if 'eqps' in metric_store.keys() and transaction in metric_store['eqps'].keys():
        transaction_key = transaction + '.' + 'ErrorsPerSecond'
        self.calculated_stats[transaction_key], self.calculated_percentiles[transaction_key] = \
            naarad.utils.calculate_stats(metric_store['eqps'][transaction].values(),
                                         stats_to_calculate, percentiles_to_calculate)
        self.update_summary_stats(transaction + '.' + 'ErrorsPerSecond')
      transaction_key = transaction + '.' + 'DataThroughput'
      self.calculated_stats[transaction_key], self.calculated_percentiles[transaction_key] = \
          naarad.utils.calculate_stats(metric_store['thr'][transaction].values(),
                                       stats_to_calculate, percentiles_to_calculate)
      self.update_summary_stats(transaction_key)
    return None

  def parse(self):
    """
    Parse the Jmeter file and calculate key stats

    :return: status of the metric parse
    """
    file_status = True
    for infile in self.infile_list:
      file_status = file_status and naarad.utils.is_valid_file(infile)
      if not file_status:
        return False

    status = self.parse_xml_jtl(self.aggregation_granularity)
    gc.collect()
    return status

  def _sanitize_label(self, raw_label):
    return raw_label.replace('/', '_').replace('?', '_')

  def parse_xml_jtl(self, granularity):
    """
    Parse Jmeter workload output in XML format and extract overall and per transaction data and key statistics

    :param string granularity: The time period over which to aggregate and average the raw data. Valid values are 'hour', 'minute' or 'second'
    :return: status of the metric parse
    """
    data = defaultdict(list)
    processed_data = defaultdict(lambda: defaultdict(lambda: defaultdict(list)))
    for input_file in self.infile_list:
      logger.info('Processing : %s', input_file)
      timestamp_format = None
      tree = ElementTree.parse(input_file)
      samples = tree.findall('./httpSample')
      for sample in samples:
        if not timestamp_format or timestamp_format == 'unknown':
          timestamp_format = naarad.utils.detect_timestamp_format(sample.get('ts'))
        if timestamp_format == 'unknown':
          continue
        ts = naarad.utils.get_standardized_timestamp(sample.get('ts'), timestamp_format)
        if ts == -1:
          continue
        ts = naarad.utils.reconcile_timezones(ts, self.timezone, self.graph_timezone)
        aggregate_timestamp, averaging_factor = self.get_aggregation_timestamp(ts, granularity)
        self.aggregate_count_over_time(processed_data, sample, [self._sanitize_label(sample.get('lb')), 'Overall_Summary'], aggregate_timestamp)
        self.aggregate_values_over_time(processed_data, sample, [self._sanitize_label(sample.get('lb')), 'Overall_Summary'], ['t', 'by'], aggregate_timestamp)
        logger.info('Finished parsing : %s', input_file)
    logger.info('Processing metrics for output to csv')
    self.average_values_for_plot(processed_data, data, averaging_factor)
    logger.info('Writing time series csv')
    for csv in data.keys():
      self.csv_files.append(csv)
      with open(csv, 'w') as csvf:
        csvf.write('\n'.join(sorted(data[csv])))
    logger.info('Processing raw data for stats')
    self.calculate_key_stats(processed_data)
    return True

  def calculate_stats(self):
    stats_csv = self.get_stats_csv()
    imp_metric_stats_csv = self.get_important_sub_metrics_csv()
    csv_header = 'sub_metric,mean,std. deviation,median,min,max,90%,95%,99%\n'
    with open(stats_csv, 'w') as FH:
      FH.write(csv_header)
      for sub_metric in self.calculated_stats:
        percentile_data = self.calculated_percentiles[sub_metric]
        stats_data = self.calculated_stats[sub_metric]
        csv_data = ','.join([sub_metric, str(round(stats_data['mean'], 2)), str(round(stats_data['std'], 2)), str(round(stats_data['median'], 2)),
                             str(round(stats_data['min'], 2)), str(round(stats_data['max'], 2)), str(round(percentile_data[90], 2)),
                             str(round(percentile_data[95], 2)), str(round(percentile_data[99], 2))])
        FH.write(csv_data + '\n')
      self.stats_files.append(stats_csv)
    for sub_metric in self.calculated_percentiles:
      percentiles_csv = self.get_csv(sub_metric, 'percentiles')
      percentile_data = self.calculated_percentiles[sub_metric]
      with open(percentiles_csv, 'w') as FH:
        for percentile in sorted(percentile_data):
          FH.write(str(percentile) + ',' + str(numpy.round_(percentile_data[percentile], 2)) + '\n')
        self.percentiles_files.append(percentiles_csv)
    with open(imp_metric_stats_csv, 'w') as FH_IMP:
      FH_IMP.write(csv_header)
      for sub_metric in self.important_sub_metrics:
        if sub_metric in self.calculated_stats.keys():
          percentile_data = self.calculated_percentiles[sub_metric]
          stats_data = self.calculated_stats[sub_metric]
          csv_data = ','.join([sub_metric, str(round(stats_data['mean'], 2)), str(round(stats_data['std'], 2)), str(round(stats_data['median'], 2)),
                               str(round(stats_data['min'], 2)), str(round(stats_data['max'], 2)), str(round(percentile_data[90], 2)),
                               str(round(percentile_data[95], 2)), str(round(percentile_data[99], 2))])
          FH_IMP.write(csv_data + '\n')
      self.important_stats_files.append(imp_metric_stats_csv)

  def plot_timeseries(self, graphing_library='matplotlib'):
    if graphing_library != 'matplotlib':
      return Metric.plot_timeseries(self, graphing_library)
    else:
      logger.info('Using graphing_library {lib} for metric {name}'.format(lib=graphing_library, name=self.label))
      plot_data = {}
      # plot time series data for submetrics
      for out_csv in sorted(self.csv_files, reverse=True):
        csv_filename = os.path.basename(out_csv)
        # The last element is .csv, don't need that in the name of the chart
        column = csv_filename.split('.')[-2]
        transaction_name = ' '.join(csv_filename.split('.')[1:-2])
        plot = PD(input_csv=out_csv, csv_column=1, series_name=transaction_name,
                  y_label=self.sub_metric_description[column] + ' (' + self.sub_metric_units[column] + ')', precision=None, graph_height=500, graph_width=1200,
                  graph_type='line')
        if transaction_name in plot_data:
          plot_data[transaction_name].append(plot)
        else:
          plot_data[transaction_name] = [plot]
      for transaction in plot_data:
        graphed, div_file = Metric.graphing_modules[graphing_library].graph_data(plot_data[transaction], self.resource_directory, self.resource_path,
                                                                                 self.label + '.' + transaction)
        if graphed:
          self.plot_files.append(div_file)
      return True



# coding=utf-8
"""
Copyright 2013 LinkedIn Corp. All rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import sys
import os
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), 'src')))
import naarad.utils
from naarad.metrics.metric import Metric
from naarad.reporting.report import Report
output_directory = ''
resource_directory = ''
input_log_directory = ''
resource_path = 'resources'


def setup_module():
  global output_directory
  global resource_path
  global resource_directory
  global input_log_directory
  output_directory = os.path.join('.', 'tmp_report_test')
  resource_directory = os.path.join(output_directory, resource_path)
  input_log_directory = os.path.join(output_directory, 'logs')
  if not os.path.exists(output_directory):
    os.makedirs(output_directory)
  if not os.path.exists(input_log_directory):
    os.makedirs(input_log_directory)
  if not os.path.exists(resource_directory):
    os.makedirs(resource_directory)
  with open(os.path.join(input_log_directory, 'a.csv'), 'w') as TestFile1:
    TestFile1.write('2014-03-26 00:00:00, 0')
  with open(os.path.join(input_log_directory, 'b.csv'), 'w') as TestFile1:
    TestFile1.write('2014-03-26 00:00:00, 1')


def get_three_metrics(output_directory, resource_path, rules):
  metrics = [Metric('MetricOne', 'TestOne.csv', 'HostnameOne', output_directory, resource_path, 'MetricOne', None, None, rules, None, None),
             Metric('MetricTwo', 'TestTwo.csv', 'HostnameOne', output_directory, resource_path, 'MetricTwo', None, None, rules, None, None),
             Metric('MetricThree', 'TestThree.csv', 'HostnameOne', output_directory, resource_path, 'MetricThree', None, None, rules, None, None)]
  return metrics


def get_two_metrics(output_directory, resource_path, rules):
  metrics = [Metric('MetricOne', 'TestOne.csv', 'HostnameOne', output_directory, resource_path, 'MetricOne', None, None, rules, None, None),
             Metric('MetricTwo', 'TestTwo.csv', 'HostnameOne', output_directory, resource_path, 'MetricTwo', None, None, rules, None, None)]
  return metrics


def test_metrics_without_summary_with_error():
  """
  Tests to verify that metric reports are not generated if the metrics are in error. Also no summary report should be created.
  """
  global output_directory
  global resource_path
  rules = {}
  metrics = get_three_metrics(output_directory, resource_path, rules)
  aggregate_metrics = []
  correlated_plots = []
  rpt = Report(None, output_directory, resource_directory, resource_path, metrics + aggregate_metrics, correlated_plots=correlated_plots)
  rpt.generate()
  assert not naarad.utils.is_valid_file(os.path.join(output_directory, 'MetricOne_report.html'))
  assert not naarad.utils.is_valid_file(os.path.join(output_directory, 'MetricTwo_report.html'))
  assert not naarad.utils.is_valid_file(os.path.join(output_directory, 'MetricThree_report.html'))
  assert naarad.utils.is_valid_file(os.path.join(output_directory, 'report.html'))
  assert not naarad.utils.is_valid_file(os.path.join(output_directory, 'summary_report.html'))
  os.system('rm -rf tmp_report_test/*.*')
  os.system('rm -rf tmp_report_test/resources/*.*')


def test_metrics_without_summary_without_error():
  """
  Tests to verify that metric reports are generated if the metrics are not in error. Also no summary report should be created.
  """
  global output_directory
  global input_log_directory
  global resource_path
  rules = {}
  metrics = get_three_metrics(output_directory, resource_path, rules)
  for metric in metrics:
    files_list = [os.path.join(input_log_directory, 'a.csv'), os.path.join(input_log_directory, 'b.csv')]
    metric.csv_files = files_list
    metric.stats_files = files_list
    metric.timeseries_csv_list = files_list
    metric.percentiles_files = files_list

  aggregate_metrics = []
  correlated_plots = []
  rpt = Report(None, output_directory, resource_directory, resource_path, metrics + aggregate_metrics, correlated_plots=correlated_plots)
  rpt.generate()
  assert naarad.utils.is_valid_file(os.path.join(output_directory, 'MetricOne_report.html'))
  assert naarad.utils.is_valid_file(os.path.join(output_directory, 'MetricTwo_report.html'))
  assert naarad.utils.is_valid_file(os.path.join(output_directory, 'MetricThree_report.html'))
  assert naarad.utils.is_valid_file(os.path.join(output_directory, 'report.html'))
  assert not naarad.utils.is_valid_file(os.path.join(output_directory, 'summary_report.html'))
  os.system('rm -rf tmp_report_test/*.*')
  os.system('rm -rf tmp_report_test/resources/*.*')


def test_metrics_with_summary_without_error():
  """
  Tests to verify that metric reports are generated if the metrics are not in error. Also a summary report should be created.
  """
  global output_directory
  global input_log_directory
  global resource_path
  rules = {}
  metrics = get_three_metrics(output_directory, resource_path, rules)
  for metric in metrics:
    files_list = [os.path.join(input_log_directory, 'a.csv'), os.path.join(input_log_directory, 'b.csv')]
    metric.csv_files = files_list
    metric.stats_files = files_list
    metric.timeseries_csv_list = files_list
    metric.important_stats_files = files_list
    metric.percentiles_files = files_list

  aggregate_metrics = []
  correlated_plots = []
  rpt = Report(None, output_directory, resource_directory, resource_path, metrics + aggregate_metrics, correlated_plots=correlated_plots)
  rpt.generate()
  assert naarad.utils.is_valid_file(os.path.join(output_directory, 'MetricOne_report.html'))
  assert naarad.utils.is_valid_file(os.path.join(output_directory, 'MetricTwo_report.html'))
  assert naarad.utils.is_valid_file(os.path.join(output_directory, 'MetricThree_report.html'))
  assert naarad.utils.is_valid_file(os.path.join(output_directory, 'report.html'))
  assert naarad.utils.is_valid_file(os.path.join(output_directory, 'summary_report.html'))
  os.system('rm -rf tmp_report_test/*.*')
  os.system('rm -rf tmp_report_test/resources/*.*')


def test_metrics_with_summary_with_partial_error():
  """
  Tests to verify that metric reports are generated for OK metrics if there are some metrics that are in error. Also a summary report should be created.
  """
  global output_directory
  global input_log_directory
  global resource_path
  rules = {}
  metrics = get_two_metrics(output_directory, resource_path, rules)
  files_list = [os.path.join(input_log_directory, 'a.csv'), os.path.join(input_log_directory, 'b.csv')]
  for metric in metrics:
    metric.csv_files = files_list
    metric.stats_files = files_list
    metric.timeseries_csv_list = files_list
    metric.important_stats_files = files_list
    metric.percentiles_files = files_list

  metrics.append(Metric('MetricThree', 'TestThree.csv', 'HostnameOne', output_directory, resource_path, 'MetricThree', None, None, rules, None, None))

  aggregate_metrics = []
  correlated_plots = []
  rpt = Report(None, output_directory, resource_directory, resource_path, metrics + aggregate_metrics, correlated_plots=correlated_plots)
  rpt.generate()
  assert naarad.utils.is_valid_file(os.path.join(output_directory, 'MetricOne_report.html'))
  assert naarad.utils.is_valid_file(os.path.join(output_directory, 'MetricTwo_report.html'))
  assert not naarad.utils.is_valid_file(os.path.join(output_directory, 'MetricThree_report.html'))
  assert naarad.utils.is_valid_file(os.path.join(output_directory, 'report.html'))
  assert naarad.utils.is_valid_file(os.path.join(output_directory, 'summary_report.html'))
  os.system('rm -rf tmp_report_test/*.*')
  os.system('rm -rf tmp_report_test/resources/*.*')


def test_metrics_without_summary_with_partial_error():
  """
  Tests to verify that metric reports are generated for OK metrics if there are some metrics that are in error. Also no summary report should be created.
  """
  global output_directory
  global input_log_directory
  global resource_path
  rules = {}
  metrics = get_two_metrics(output_directory, resource_path, rules)
  files_list = [os.path.join(input_log_directory, 'a.csv'), os.path.join(input_log_directory, 'b.csv')]
  for metric in metrics:
    metric.csv_files = files_list
    metric.stats_files = files_list
    metric.timeseries_csv_list = files_list
    metric.percentiles_files = files_list

  metrics.append(Metric('MetricThree', 'TestThree.csv', 'HostnameOne', output_directory, resource_path, 'MetricThree', None, None, rules, None, None))

  aggregate_metrics = []
  correlated_plots = []
  rpt = Report(None, output_directory, resource_directory, resource_path, metrics + aggregate_metrics, correlated_plots=correlated_plots)
  rpt.generate()
  assert naarad.utils.is_valid_file(os.path.join(output_directory, 'MetricOne_report.html'))
  assert naarad.utils.is_valid_file(os.path.join(output_directory, 'MetricTwo_report.html'))
  assert not naarad.utils.is_valid_file(os.path.join(output_directory, 'MetricThree_report.html'))
  assert naarad.utils.is_valid_file(os.path.join(output_directory, 'report.html'))
  assert not naarad.utils.is_valid_file(os.path.join(output_directory, 'summary_report.html'))
  os.system('rm -rf tmp_report_test/*.*')
  os.system('rm -rf tmp_report_test/resources/*.*')


def tearDown():
  os.system('rm -rf tmp_report_test')



# coding=utf-8
"""
Copyright 2013 LinkedIn Corp. All rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os.path
import csv
import errno
from jinja2 import Environment, FileSystemLoader
import logging
import os
import shutil
from collections import defaultdict
import naarad.httpdownload
import naarad.utils
import naarad.naarad_constants as CONSTANTS
import naarad.resources
from naarad.graphing.plot_data import PlotData as PD
import ConfigParser
import time

logger = logging.getLogger('naarad.reporting.diff')


class Diff(object):
  """
  Support for diff between two naarad reports is provided as a diff report implemented by this class. The two reports to
  be diffed can be local or remote (http(s)) on same or different hosts. Metrics and stats in common between the two
  reports are diffed both as absolute as well as %. Client charting is supported for time series in common between the 2
  reports. Metrics & Statistics that are not in common between the two reports are ignored.
  """

  def __init__(self, reports_list, report_name, output_directory, resource_directory, resource_path, **other_options):
    self.reports = reports_list
    if report_name == '':
      self.report_name = 'diff'
    else:
      self.report_name = report_name
    self.output_directory = output_directory
    self.resource_directory = resource_directory
    self.resource_path = resource_path
    if other_options:
      for (key, val) in other_options.iteritems():
        setattr(self, key, val)
    self.status = 'OK'
    self.stylesheet_includes = CONSTANTS.STYLESHEET_INCLUDES
    self.javascript_includes = CONSTANTS.JAVASCRIPT_INCLUDES
    self.diff_data = defaultdict(lambda: defaultdict(lambda: defaultdict(dict)))
    self.plot_files = []
    self.sla_map = defaultdict(lambda: defaultdict(lambda: defaultdict(None)))
    self.sla_failures = 0
    self.sla_failure_list = []

  def get_resources_location(self):
    """
    Return the location of the report templates and local js/css includes.
    :return: filesystem location for naarad/src/resources directory.
    """
    return naarad.resources.get_dir()

  def copy_local_includes(self):
    """
    Copy local js/css includes from naarad resources to the report/resources directory
    :return: None
    """
    resource_folder = self.get_resources_location()
    for stylesheet in self.stylesheet_includes:
      if ('http' not in stylesheet) and naarad.utils.is_valid_file(os.path.join(resource_folder, stylesheet)):
        shutil.copy(os.path.join(resource_folder, stylesheet), self.resource_directory)

    for javascript in self.javascript_includes:
      if ('http' not in javascript) and naarad.utils.is_valid_file(os.path.join(resource_folder, javascript)):
        shutil.copy(os.path.join(resource_folder, javascript), self.resource_directory)

    return None

  def generate_client_charting_page(self, data_sources):
    """
    Create the client charting page for the diff report, with time series data from the two diffed reports.
    :return: generated html to be written to disk
    """
    if not os.path.exists(self.resource_directory):
      os.makedirs(self.resource_directory)
    self.copy_local_includes()
    template_loader = FileSystemLoader(self.get_resources_location())
    template_environment = Environment(loader=template_loader)
    client_html = template_environment.get_template(CONSTANTS.TEMPLATE_HEADER).render(custom_stylesheet_includes=CONSTANTS.STYLESHEET_INCLUDES,
                                                                                      custom_javascript_includes=CONSTANTS.JAVASCRIPT_INCLUDES,
                                                                                      resource_path=self.resource_path,
                                                                                      report_title='naarad diff report') + '\n'
    client_html += template_environment.get_template(CONSTANTS.TEMPLATE_DIFF_CLIENT_CHARTING).render(data_series=data_sources,
                                                                                                     resource_path=self.resource_path) + '\n'
    client_html += template_environment.get_template(CONSTANTS.TEMPLATE_FOOTER).render()
    return client_html

  def generate_diff_html(self):
    """
    Generate the summary diff report html from template
    :return: generated html to be written to disk
    """
    if not os.path.exists(self.resource_directory):
      os.makedirs(self.resource_directory)
    self.copy_local_includes()
    div_html = ''
    for plot_div in sorted(self.plot_files):
      with open(plot_div, 'r') as div_file:
        div_html += '\n' + div_file.read()
    template_loader = FileSystemLoader(self.get_resources_location())
    template_environment = Environment(loader=template_loader)
    template_environment.filters['sanitize_string'] = naarad.utils.sanitize_string
    diff_html = template_environment.get_template(CONSTANTS.TEMPLATE_HEADER).render(custom_stylesheet_includes=CONSTANTS.STYLESHEET_INCLUDES,
                                                                                    custom_javascript_includes=CONSTANTS.JAVASCRIPT_INCLUDES,
                                                                                    resource_path=self.resource_path,
                                                                                    report_title='naarad diff report') + '\n'
    diff_html += template_environment.get_template(CONSTANTS.TEMPLATE_DIFF_PAGE).render(diff_data=self.diff_data, plot_div_content=div_html,
                                                                                        reports=self.reports, sla_failure_list=self.sla_failure_list,
                                                                                        sla_map=self.sla_map) + '\n'
    diff_html += template_environment.get_template(CONSTANTS.TEMPLATE_FOOTER).render()
    return diff_html

  def discover(self, metafile):
    """
    Determine what summary stats, time series, and CDF csv exist for the reports that need to be diffed.
    :return: boolean: return whether the summary stats / time series / CDF csv summary was successfully located
    """
    for report in self.reports:
      if report.remote_location == 'local':
        if naarad.utils.is_valid_file(os.path.join(os.path.join(report.location, self.resource_path), metafile)):
          with open(os.path.join(os.path.join(report.location, self.resource_path), metafile), 'r') as meta_file:
            if metafile == CONSTANTS.STATS_CSV_LIST_FILE:
              report.stats = meta_file.readlines()[0].split(',')
            elif metafile == CONSTANTS.PLOTS_CSV_LIST_FILE:
              report.datasource = meta_file.readlines()[0].split(',')
            elif metafile == CONSTANTS.CDF_PLOTS_CSV_LIST_FILE:
              report.cdf_datasource = meta_file.readlines()[0].split(',')
        else:
            report.status = 'NO_SUMMARY_STATS'
            self.status = 'ERROR'
            logger.error('Unable to access summary stats file for report :%s', report.label)
            return False
      else:
        stats_url = report.remote_location + '/' + self.resource_path + '/' + metafile
        meta_file_data = naarad.httpdownload.stream_url(stats_url)

        if meta_file_data:
          if metafile == CONSTANTS.STATS_CSV_LIST_FILE:
            report.stats = meta_file_data.split(',')
          elif metafile == CONSTANTS.PLOTS_CSV_LIST_FILE:
            report.datasource = meta_file_data.split(',')
          elif metafile == CONSTANTS.CDF_PLOTS_CSV_LIST_FILE:
            report.cdf_datasource = meta_file_data.split(',')
        else:
          report.status = 'NO_SUMMARY_STATS'
          self.status = 'ERROR'
          logger.error('No summary stats available for report :%s', report.label)
          return False
    return True

  def collect_datasources(self):
    """
    Identify what time series exist in both the diffed reports and download them to the diff report resources directory
    :return: True/False : return status of whether the download of time series resources succeeded.
    """
    report_count = 0
    if self.status != 'OK':
      return False
    diff_datasource = sorted(set(self.reports[0].datasource) & set(self.reports[1].datasource))
    if diff_datasource:
      self.reports[0].datasource = diff_datasource
      self.reports[1].datasource = diff_datasource
    else:
      self.status = 'NO_COMMON_STATS'
      logger.error('No common metrics were found between the two reports')
      return False
    for report in self.reports:
      report.label = report_count
      report_count += 1
      report.local_location = os.path.join(self.resource_directory, str(report.label))
      try:
        os.makedirs(report.local_location)
      except OSError as exeption:
        if exeption.errno != errno.EEXIST:
          raise
      if report.remote_location != 'local':
        naarad.httpdownload.download_url_list(map(lambda x: report.remote_location + '/' + self.resource_path + '/' + x + '.csv', report.datasource),
                                              report.local_location)
      else:
        for filename in report.datasource:
          try:
            shutil.copy(os.path.join(os.path.join(report.location, self.resource_path), filename + '.csv'), report.local_location)
          except IOError as exeption:
            continue
    return True

  def collect_cdf_datasources(self):
    """
    Identify what cdf series exist in both the diffed reports and download them to the diff report resources directory
    :return: True/False : return status of whether the download of time series resources succeeded.
    """
    report_count = 0
    if self.status != 'OK':
      return False
    diff_cdf_datasource = sorted(set(self.reports[0].cdf_datasource) & set(self.reports[1].cdf_datasource))
    if diff_cdf_datasource:
      self.reports[0].cdf_datasource = diff_cdf_datasource
      self.reports[1].cdf_datasource = diff_cdf_datasource
    else:
      self.status = 'NO_COMMON_STATS'
      logger.error('No common metrics were found between the two reports')
      return False
    for report in self.reports:
      report.label = report_count
      report_count += 1
      report.local_location = os.path.join(self.resource_directory, str(report.label))
      try:
        os.makedirs(report.local_location)
      except OSError as exeption:
        if exeption.errno != errno.EEXIST:
          raise
      if report.remote_location != 'local':
        naarad.httpdownload.download_url_list(map(lambda x: report.remote_location + '/' + self.resource_path + '/' + x + '.csv', report.cdf_datasource),
                                              report.local_location)
      else:
        for filename in report.cdf_datasource:
          try:
            shutil.copy(os.path.join(os.path.join(report.location, self.resource_path), filename + '.csv'), report.local_location)
          except IOError as exeption:
            continue
    return True

  def collect(self):
    """
    Identify what summary stats exist in both the diffed reports and download them to the diff report resources directory
    :return: True/False : return status of whether the download of summary stats succeeded.
    """
    report_count = 0
    if self.status != 'OK':
      return False
    diff_stats = set(self.reports[0].stats) & set(self.reports[1].stats)
    if diff_stats:
      self.reports[0].stats = diff_stats
      self.reports[1].stats = diff_stats
    else:
      self.status = 'NO_COMMON_STATS'
      logger.error('No common metrics were found between the two reports')
      return False
    for report in self.reports:
      report.label = report_count
      report_count += 1
      report.local_location = os.path.join(self.resource_directory, str(report.label))
      try:
        os.makedirs(report.local_location)
      except OSError as exeption:
        if exeption.errno != errno.EEXIST:
          raise
      if report.remote_location != 'local':
        naarad.httpdownload.download_url_list(map(lambda x: report.remote_location + '/' + self.resource_path + '/' + x, report.stats), report.local_location)
      else:
          for filename in report.stats:
            shutil.copy(os.path.join(os.path.join(report.location, self.resource_path), filename), report.local_location)
    return True

  def plot_diff(self, graphing_library='matplotlib'):
    """
    Generate CDF diff plots of the submetrics
    """
    diff_datasource = sorted(set(self.reports[0].datasource) & set(self.reports[1].datasource))
    graphed = False
    for submetric in diff_datasource:
      baseline_csv = naarad.utils.get_default_csv(self.reports[0].local_location, (submetric + '.percentiles'))
      current_csv = naarad.utils.get_default_csv(self.reports[1].local_location, (submetric + '.percentiles'))
      if (not (naarad.utils.is_valid_file(baseline_csv) & naarad.utils.is_valid_file(current_csv))):
        continue
      baseline_plot = PD(input_csv=baseline_csv, csv_column=1, series_name=submetric, y_label=submetric, precision=None, graph_height=600, graph_width=1200,
                         graph_type='line', plot_label='baseline', x_label='Percentiles')
      current_plot = PD(input_csv=current_csv, csv_column=1, series_name=submetric, y_label=submetric, precision=None, graph_height=600, graph_width=1200,
                        graph_type='line', plot_label='current', x_label='Percentiles')
      graphed, div_file = Diff.graphing_modules[graphing_library].graph_data_on_the_same_graph([baseline_plot, current_plot],
                                                                                               os.path.join(self.output_directory, self.resource_path),
                                                                                               self.resource_path, (submetric + '.diff'))
      if graphed:
        self.plot_files.append(div_file)
    return True

  def check_sla(self, sla, diff_metric):
    """
    Check whether the SLA has passed or failed
    """
    try:
      if sla.display is '%':
        diff_val = float(diff_metric['percent_diff'])
      else:
        diff_val = float(diff_metric['absolute_diff'])
    except ValueError:
      return False
    if not (sla.check_sla_passed(diff_val)):
      self.sla_failures += 1
      self.sla_failure_list.append(DiffSLAFailure(sla, diff_metric))
    return True

  def generate(self):
    """
    Generate a diff report from the reports specified.
    :return: True/False : return status of whether the diff report generation succeeded.
    """
    if (self.discover(CONSTANTS.STATS_CSV_LIST_FILE) and self.discover(CONSTANTS.PLOTS_CSV_LIST_FILE) and self.discover(CONSTANTS.CDF_PLOTS_CSV_LIST_FILE) and
       self.collect() and self.collect_datasources() and self.collect_cdf_datasources()):
      for stats in self.reports[0].stats:
        metric_label = stats.replace('.stats.csv', '')
        stats_0 = os.path.join(self.reports[0].local_location, stats)
        stats_1 = os.path.join(self.reports[1].local_location, stats)
        report0_stats = {}
        report1_stats = {}
        if naarad.utils.is_valid_file(stats_0) and naarad.utils.is_valid_file(stats_1):
          report0 = csv.DictReader(open(stats_0))
          for row in report0:
            report0_stats[row[CONSTANTS.SUBMETRIC_HEADER]] = row
          report0_stats['__headers__'] = report0._fieldnames
          report1 = csv.DictReader(open(stats_1))
          for row in report1:
            report1_stats[row[CONSTANTS.SUBMETRIC_HEADER]] = row
          report1_stats['__headers__'] = report1._fieldnames
          common_stats = sorted(set(report0_stats['__headers__']) & set(report1_stats['__headers__']))
          common_submetrics = sorted(set(report0_stats.keys()) & set(report1_stats.keys()))
          for submetric in common_submetrics:
            if submetric != '__headers__':
              for stat in common_stats:
                if stat != CONSTANTS.SUBMETRIC_HEADER:
                  diff_metric = reduce(defaultdict.__getitem__, [stats.split('.')[0], submetric, stat], self.diff_data)
                  diff_metric[0] = float(report0_stats[submetric][stat])
                  diff_metric[1] = float(report1_stats[submetric][stat])
                  diff_metric['absolute_diff'] = naarad.utils.normalize_float_for_display(diff_metric[1] - diff_metric[0])
                  if diff_metric[0] == 0:
                    if diff_metric['absolute_diff'] == '0.0':
                      diff_metric['percent_diff'] = 0.0
                    else:
                      diff_metric['percent_diff'] = 'N/A'
                  else:
                    diff_metric['percent_diff'] = naarad.utils.normalize_float_for_display((diff_metric[1] - diff_metric[0]) * 100 / diff_metric[0])
                  # check whether there is a SLA failure
                  if ((metric_label in self.sla_map.keys()) and (submetric in self.sla_map[metric_label].keys()) and
                     (stat in self.sla_map[metric_label][submetric].keys())):
                    self.check_sla(self.sla_map[metric_label][submetric][stat], diff_metric)
    else:
      return False
    self.plot_diff()
    diff_html = ''
    if self.diff_data:
      diff_html = self.generate_diff_html()
      client_html = self.generate_client_charting_page(self.reports[0].datasource)
    if diff_html != '':
      with open(os.path.join(self.output_directory, CONSTANTS.DIFF_REPORT_FILE), 'w') as diff_file:
        diff_file.write(diff_html)
      with open(os.path.join(self.output_directory, CONSTANTS.CLIENT_CHARTING_FILE), 'w') as client_file:
        client_file.write(client_html)
    return True


class DiffSLAFailure:
  """
  DiffSLAFailures object tracks the information related to a SLA failure
  """
  def __init__(self, sla, diff_metric):
    self.sla = sla
    self.diff_metric = diff_metric


class NaaradReport:
  """
  NaaradReport object contains details on the the reports being diffed, such as details on available summary stats,
  timeseries and location in current diff report where the collected data is stored.
  """
  def __init__(self, location, title):
    if location.startswith('http://') or location.startswith('https://'):
      self.remote_location = location
    else:
      self.remote_location = 'local'
      self.local_location = location
    self.location = location
    if title == '':
      self.title = location
    self.status = 'OK'
    self.stats = []
    self.datasource = []
    self.cdf_datasource = []
    self.label = ''



# coding=utf-8
"""
Copyright 2013 LinkedIn Corp. All rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import datetime
import logging
import os
import re

from naarad.metrics.metric import Metric
import naarad.utils
from naarad.naarad_constants import important_sub_metrics_import

logger = logging.getLogger('naarad.metrics.ProcInterruptsMetric')


class ProcInterruptsMetric(Metric):
  """
  Metric parser for /proc/interrupts log file.
  This parser will parse interrupts per core and per IRQ and store the differences between intervals. Thus the first entry will be considered 0.
  This is because the interrupts are an increasing counter rather than per interval.

  Extra Configurations:

    You can specify CPUS= in the configuration file to define which CPUs to look at. For example to have the CPUs 0,
    3, 4, and 8 then specify the following line in the PROCINTERRUPTS section of the configurations:

    [PROCINTERRUPTS]
    CPUS=CPU0 CPU3 CPU4 CPU8
    ...

  You can of course use the other general metric settings such as ts_start, infile, etc.
  """
  def __init__(self, metric_type, infile_list, hostname, aggr_metrics, outdir, resource_path, label, ts_start, ts_end, rule_strings,
               important_sub_metrics, anomaly_detection_metrics, **other_options):
    Metric.__init__(self, metric_type, infile_list,  hostname, aggr_metrics, outdir, resource_path, label, ts_start, ts_end, rule_strings,
                    important_sub_metrics, anomaly_detection_metrics)
    if not self.important_sub_metrics and self.metric_type in important_sub_metrics_import.keys():
      self.important_sub_metrics = important_sub_metrics_import[self.metric_type]
    self.options = None

    self.CPUS = None
    for (key, val) in other_options.iteritems():
      setattr(self, key, val.split())

  def get_csv(self, cpu, device=None):
    """
    Returns the CSV file related to the given metric. The metric is determined by the cpu and device.
    The cpu is the CPU as in the interrupts file for example CPU12.
    The metric is a combination of the CPU and device. The device consists of IRQ #, the irq device ASCII name.

                                      CPU0   CPU1
    2014-10-29 00:27:42.15161    59:    29      2    IR-IO-APIC-edge    timer
                                  ^      ^      ^          ^              ^
                                  |      |      |          |              |
                                IRQ#   Value  Value   IRQ Device       Ascii Name

    This would produce a metric CPU0.timer-IRQ59 and CPU1.timer-IRQ59 so one per IRQ per CPU.

    :param cpu: The name of the cpu given as CPU#.
    :param device: The device name as given by the system. <ASCII name>-IRQ<IRQ #>
    :return: The CSV file for the metric.
    """
    cpu = naarad.utils.sanitize_string(cpu)
    if device is None:
      outcsv = os.path.join(self.resource_directory, "{0}.{1}.csv".format(self.label, cpu))
      self.csv_column_map[outcsv] = cpu
    else:
      device = naarad.utils.sanitize_string(device)
      outcsv = os.path.join(self.resource_directory, "{0}.{1}.{2}.csv".format(self.label, cpu, device))
      self.csv_column_map[outcsv] = cpu + '.' + device
    return outcsv

  def is_header_line(self, line):
    """
    Checks to see if the line is a header line.
    The header line style is currently:

      2014-10-29 00:28:42.15161        CPU0   CPU1   CPU2   CPU3  ...

    :param line: The line of the file to check.
    :return: Boolean of true or false of whether line is header line or not.
    """
    return 'CPU' in line

  def find_header(self, infile):
    """
    Parses the file and tries to find the header line. The header line has format:

      2014-10-29 00:28:42.15161        CPU0   CPU1   CPU2   CPU3  ...

    So should always have CPU# for each core. This function verifies a good header and
    returns the list of CPUs that exist from the header.

    :param infile: The opened file in read mode to find the header.
    :return cpus: A list of the core names so in this example ['CPU0', 'CPU1', ...]
    """
    cpus = []
    for line in infile:  # Pre-processing - Try to find header
      if not self.is_header_line(line):
        continue
      # Verifying correctness of the header
      cpu_header = line.split()
      for cpu_h in cpu_header[2:]:
        if not cpu_h.startswith('CPU'):
          cpus = []  # Bad header so reset to nothing
          break
        else:
          cpus.append(cpu_h)
      if len(cpus) > 0:  # We found the header
        break
    return cpus

  def parse(self):
    """
    Processes the files for each IRQ and each CPU in terms of the differences.
    Also produces accumulated interrupt count differences for each set of Ethernet IRQs.
    Generally Ethernet has 8 TxRx IRQs thus all are combined so that one can see the overall interrupts being generated by the NIC.

    Simplified Interrupt File Format: (See examples for example log)

                                        CPU0   CPU1
      2014-10-29 00:27:42.15161    59:    29      2    IR-IO-APIC-edge    timer
      2014-10-29 00:27:42.15161    60:  2123      0    IR-PCI-MSI-edge    eth0

                                        CPU0   CPU1
      2014-10-29 00:27:42.15161    59:    29      2    IR-IO-APIC-edge    timer
      2014-10-29 00:27:42.15161    60:  2123      0    IR-PCI-MSI-edge    eth0

    :returns: True or False whether parsing was successful or not.
    """
    if not os.path.isdir(self.outdir):
      os.makedirs(self.outdir)
    if not os.path.isdir(self.resource_directory):
      os.makedirs(self.resource_directory)

    data = {}
    for input_file in self.infile_list:
      logger.info('Processing : %s', input_file)
      timestamp_format = None
      with open(input_file, 'r') as infile:
        # Get the header for this file
        cpus = self.find_header(infile)
        if len(cpus) == 0:  # Make sure we have header otherwise go to next file
          logger.error("Header not found for file: %s", input_file)
          continue

        # Parse the actual file after header
        prev_data = None    # Stores the previous interval's log data
        curr_data = {}      # Stores the current interval's log data
        eth_data = {}
        for line in infile:
          if self.is_header_line(line):  # New section so save old and aggregate ETH
            prev_data = curr_data
            curr_data = {}
            # New section so store the collected Ethernet data
            # Example Aggregate metric: PROCINTERRUPTS.AGGREGATE.eth0
            for eth in eth_data:
              outcsv = self.get_csv('AGGREGATE', eth)
              if outcsv not in data:
                data[outcsv] = []
              data[outcsv].append(ts + ',' + str(eth_data[eth]))
            eth_data = {}
            continue

          words = line.split()
          if len(words) <= 4:  # Does not have any CPU data so skip
            continue

          # Process timestamp or determine timestamp
          ts = words[0] + " " + words[1]
          if not timestamp_format or timestamp_format == 'unknown':
            timestamp_format = naarad.utils.detect_timestamp_format(ts)
          if timestamp_format == 'unknown':
            continue
          ts = naarad.utils.get_standardized_timestamp(ts, timestamp_format)
          if self.ts_out_of_range(ts):  # See if time is in range
            continue

          # Process data lines
          # Note that some IRQs such as ERR and MIS do not have device nor ascii name
          device = words[2].strip(':')  # Get IRQ Number/Name
          if re.match("\d+", device):
            # Devices with digits need ASCII name if exists
            if (4 + len(cpus)) < len(words):
              device = words[4 + len(cpus)] + "-IRQ" + device
            else:
              device = "IRQ" + device
          else:
            # For devices with IRQ # that aren't digits then has description
            device = "-".join(words[(3 + len(cpus)):]) + "-IRQ" + device

          # Deal with each column worth of data
          for (cpu, datum) in zip(cpus, words[3:]):
            if self.CPUS and cpu not in self.CPUS:  # Skip if config defines which CPUs to look at
              continue
            outcsv = self.get_csv(cpu, device)
            curr_data[outcsv] = int(datum)
            if outcsv in data:
              datum = int(datum) - prev_data[outcsv]  # prev_data exists since outcsv exists in data
            else:
              data[outcsv] = []
              datum = 0  # First data point is set to 0
            # Store data point
            data[outcsv].append(ts + ',' + str(datum))

            # Deal with accumulating aggregate data for Ethernet
            m = re.search("(?P<eth>eth\d)", device)
            if m:
              eth = m.group('eth')
              if eth not in eth_data:
                eth_data[eth] = 0
              eth_data[eth] += datum

    # Post processing, putting data in csv files
    for csv in data.keys():
      self.csv_files.append(csv)
      with open(csv, 'w') as csvf:
        csvf.write('\n'.join(sorted(data[csv])))
    return True



# coding=utf-8
"""
Copyright 2013 LinkedIn Corp. All rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import nose
import sys
import uuid
import shutil
import time
import sys

# add the path of ~/naarad/src;   the testing py is under ~/naarad/test
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), 'src')))
import naarad.utils
from naarad.metrics.sar_metric import SARMetric
from naarad.metrics.cluster_metric import ClusterMetric

# the temporary directory for testing, will remove it after done.
tmp_dir = ''


def setup():
  create_tmp_dir()


def teardown():
  delete_tmp_dir()


def create_tmp_dir():
  '''create a unique tmp dir to hold the downloaded local files'''
  ''' if the tmp dir grenerated already exists, then simply return'''
  ''' the user simply try again to generate another unique tmp dir'''
  global tmp_dir
  tmp_dir = os.path.join('./', 'tmp' + '.' + str(uuid.uuid4()))   # ./tmp.'randomstring'
  if not os.path.exists(tmp_dir):
    os.makedirs(tmp_dir)
  else:
    print "the path of %s already exists, please try again." % tmp_dir
    return


def delete_tmp_dir():
  '''delete the tmp directory'''
  global tmp_dir
  shutil.rmtree(tmp_dir)


def test_clustermetric():
  # construct 2 SARMetric
  metric1 = SARMetric('SAR-cpuusage-host1', 'sar.cpuusage.out', 'host1', '.', 'logs', 'SAR-cpuusage-host1', None, None, {}, None, None)
  metric1.csv_column_map['logs/SAR-cpuusage-host1.all.percent-sys.csv'] = 'all.%sys'
  metric1.column_csv_map['all.%sys'] = 'logs/SAR-cpuusage-host1.all.percent-sys.csv'

  metric2 = SARMetric('SAR-cpuusage-host2', 'sar.cpuusage.out', 'host2', '.', 'logs', 'SAR-cpuusage-host2', None, None, {}, None, None)
  metric2.csv_column_map['logs/SAR-cpuusage-host2.all.percent-sys.csv'] = 'all.%sys'
  metric2.column_csv_map['all.%sys'] = 'logs/SAR-cpuusage-host2.all.percent-sys.csv'

  # construct a ClusterMetric
  aggregate_metrics = 'SAR-cpuusage.all.percent-sys:raw,avg,sum,count'
  section = 'CLUSTER-cpuusage-1'
  label = 'CLUSTER-cpuusage-1'
  resource_path = 'resources'
  rule_strings = {}
  output_directory = tmp_dir
  aggregate_hosts = 'host1 host2'
  other_options = {}
  ts_start = None
  ts_end = None
  metrics = [metric1, metric2]

  cur_metric = ClusterMetric(section, aggregate_hosts, aggregate_metrics, metrics, output_directory, resource_path, label, ts_start, ts_end, rule_strings,
                             None, None)

  # create sub-directory of resource_path
  sub_dir = os.path.join(output_directory, resource_path)
  if not os.path.exists(sub_dir):
    os.makedirs(sub_dir)

  # the only method to test; it will write to the directory the final csv files;
  cur_metric.collect()

  # check the existance of the output files
  functions = aggregate_metrics.split(':')
  prefix = functions[0].split('.')  # 'SAR-cpuusage.all.percent-sys'
  prefix[0] = section
  prefix = '.'.join(prefix)  # CLUSTER-cpuusage-1.all.percent-sys

  for func in functions[1].split(','):  # 'raw,avg,sum,count'
    file_name = prefix + '.' + func + '.csv'
    file_path = os.path.join(sub_dir, file_name)
    # print 'file to check = ' + file_path  # resources/CLUSTER-cpuusage-1.all.percent-sys.raw.csv
    assert os.path.exists(file_path)

if __name__ == '__main__':
  test_clustermetric()



# coding=utf-8
"""
Copyright 2013 LinkedIn Corp. All rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


class PlotData:
  """Class to hold details of the metrics to be plotted"""
  def __init__(self, input_csv, csv_column, series_name, y_label, precision, graph_height, graph_width, graph_type, x_label=None, plot_label=None,
               highlight_regions=None):
    self.input_csv = input_csv
    self.csv_column = csv_column
    self.graph_title = series_name
    self.y_label = y_label
    self.precision = precision
    if graph_height is None:
      self.graph_height = 600
    else:
      self.graph_height = graph_height
    if graph_width is None:
      self.graph_width = 1200
    else:
      self.graph_width = graph_width
    self.graph_type = graph_type
    self.plot_label = plot_label
    self.x_label = x_label
    self.highlight_regions = highlight_regions
    return None
